"""
Review status for each report

Revision ID: 75ae226b5d88
Revises:     fb356f0eefed
Create Date: 2022-01-27 15:19:48.185835
"""

from collections import defaultdict
from io import StringIO
import zlib

from alembic import op
import sqlalchemy as sa

from codechecker_common.source_code_comment_handler import \
    SourceCodeCommentHandler, SpellException


# Revision identifiers, used by Alembic.
revision = '75ae226b5d88'
down_revision = 'fb356f0eefed'
branch_labels = None
depends_on = None


def upgrade():
    def decode_file_content(content):
        return zlib.decompress(content).decode('utf-8', errors='ignore')

    dialect = op.get_context().dialect.name

    col_rs = sa.Column('review_status', sa.Enum(
        'unreviewed',
        'confirmed',
        'false_positive',
        'intentional',
        name='review_status'), server_default='unreviewed', nullable=True)
    col_rs_author = sa.Column(
        'review_status_author', sa.String(), nullable=True)
    col_rs_date = sa.Column(
        'review_status_date', sa.DateTime(), nullable=True)
    col_rs_is_in_source = sa.Column(
        'review_status_is_in_source',
        sa.Boolean(), server_default='0', nullable=True)
    col_rs_message = sa.Column(
        'review_status_message', sa.Binary(), nullable=True)

    conn = op.get_bind()

    if dialect == 'sqlite':
        op.execute('PRAGMA foreign_keys=off')
        with op.batch_alter_table('reports') as batch_op:
            batch_op.add_column(col_rs)
            batch_op.add_column(col_rs_author)
            batch_op.add_column(col_rs_date)
            batch_op.add_column(col_rs_is_in_source)
            batch_op.add_column(col_rs_message)
        op.execute('PRAGMA foreign_keys=on')

        # FIXME: "UPDATE ... SET ... FROM ..." is only supported starting
        # with SQLite version 3.33.0 (2020-08-14). Until this version
        # reaches LTS maturity (Ubuntu 20.04 LTS comes with 3.31.0, raising
        # a syntax error on the "FROM" in the "UPDATE" query), this
        # branching here needs to stay.
        conn.execute("""
            UPDATE reports
            SET (review_status,
                 review_status_author,
                 review_status_date,
                 review_status_message) =
                (SELECT status, author, date, message
                    FROM review_statuses
                    WHERE bug_hash = reports.bug_id)
        """)
    elif dialect == 'postgresql':
        op.add_column('reports', col_rs)
        op.add_column('reports', col_rs_author)
        op.add_column('reports', col_rs_date)
        op.add_column('reports', col_rs_is_in_source)
        op.add_column('reports', col_rs_message)

        conn.execute("""
            UPDATE reports
            SET review_status = rs.status,
                review_status_author = rs.author,
                review_status_date = rs.date,
                review_status_message = rs.message
            FROM review_statuses AS rs
            WHERE bug_id = rs.bug_hash
        """)

    conn.execute("""
        UPDATE reports
        SET review_status = 'unreviewed'
        WHERE review_status IS NULL
    """)

    files_with_report = conn.execute("""
        SELECT DISTINCT reports.file_id, files.content_hash
        FROM reports INNER JOIN files ON reports.file_id = files.id
        WHERE review_status != 'unreviewed'
    """)

    content_hashes = set()
    hash_to_content = {}
    file_id_to_content_hash = {}

    for f in files_with_report:
        content_hashes.add(f"'{f.content_hash}'")
        file_id_to_content_hash[f.file_id] = f.content_hash

    if content_hashes:
        hash_to_content = conn.execute(f"""
            SELECT content_hash, content FROM file_contents
            WHERE content_hash IN ({','.join(content_hashes)})
        """)

    hash_to_content = {
        x.content_hash: decode_file_content(x.content)
        for x in hash_to_content}

    report_id_to_line = conn.execute(f"""
        SELECT id, file_id, bug_id, checker_id, line FROM reports
        WHERE review_status != 'unreviewed'
    """)

    scch = SourceCodeCommentHandler()
    comment_cache = {}
    bug_hashes = set()
    review_status_in_source = set()
    review_status_to_report_ids = defaultdict(set)

    for row in report_id_to_line:
        cache_key = (file_id_to_content_hash[row.file_id], row.line)
        if cache_key not in comment_cache:
            try:
                comment_cache[cache_key] = scch.get_source_line_comments(
                    StringIO(
                        hash_to_content[file_id_to_content_hash[row.file_id]]),
                    row.line)
            except SpellException:
                comment_cache[cache_key] = []

        for comment in comment_cache[cache_key]:
            if row.checker_id in comment.checkers or \
                    'all' in comment.checkers:
                review_status_in_source.add(row.id)
                bug_hashes.add(f"'{row.bug_id}'")
                review_status_to_report_ids[comment.status].add(row.id)

    if review_status_in_source:
        conn.execute(f"""
            UPDATE reports
            SET review_status_is_in_source = '1'
            WHERE id IN ({','.join(map(str, review_status_in_source))})
        """)

    # Earlier a common review status belonged to all reports sharing the same
    # bug hash even if these reports had different review status given in
    # source code comment. Now these are set individually.
    for review_status, report_ids in review_status_to_report_ids.items():
        conn.execute(f"""
            UPDATE reports
            SET review_status = '{review_status}'
            WHERE id IN ({','.join(map(str, report_ids))})
        """)

    results = conn.execute("""
        SELECT bug_hash, date
        FROM review_statuses
        WHERE status IN ('false_positive', 'intentional')
    """)

    for row in results:
        if dialect == 'sqlite':
            conn.execute(f"""
                UPDATE reports
                SET fixed_at = max('{row.date}', detected_at)
                WHERE fixed_at IS NULL AND bug_id = '{row.bug_hash}'
            """)
        elif dialect == 'postgresql':
            conn.execute(f"""
                UPDATE reports
                SET fixed_at = greatest('{row.date}', detected_at)::timestamp
                WHERE fixed_at IS NULL AND bug_id = '{row.bug_hash}'
            """)

    if bug_hashes:
        conn.execute(f"""
            DELETE FROM review_statuses
            WHERE bug_hash IN ({','.join(bug_hashes)})
        """)


def downgrade():
    op.drop_column('reports', 'review_status_message')
    op.drop_column('reports', 'review_status_is_in_source')
    op.drop_column('reports', 'review_status_date')
    op.drop_column('reports', 'review_status_author')
    op.drop_column('reports', 'review_status')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Calculates call statistics from analysis output
"""

from codechecker_common.logger import get_logger

from codechecker_statistics_collector.collectors.special_return_value import \
    SpecialReturnValueCollector
from codechecker_statistics_collector.collectors.return_value import \
    ReturnValueCollector

from .analyzer import ClangSA
from ..flag import has_flag
from ..flag import prepend_all

LOG = get_logger('analyzer')


def build_stat_coll_cmd(action, config, source):
    """
    Build the statistics collector analysis command.
    """

    cmd = [ClangSA.analyzer_binary(), '-c', '-x', action.lang, '--analyze',
           # Do not warn about the unused gcc/g++ arguments.
           '-Qunused-arguments',
           '--analyzer-output', 'text']

    for plugin in ClangSA.analyzer_plugins():
        cmd.extend(["-Xclang", "-plugin",
                    "-Xclang", "checkercfg",
                    "-Xclang", "-load",
                    "-Xclang", plugin])

    cmd.extend(['-Xclang',
                '-analyzer-opt-analyze-headers'])

    cmd.extend(config.analyzer_extra_arguments)
    cmd.extend(action.analyzer_options)

    # Enable the statistics collector checkers only.
    collector_checkers = []
    checks = ClangSA.get_analyzer_checkers(alpha=True, debug=True)
    for checker_name, _ in checks:
        if SpecialReturnValueCollector.checker_collect in checker_name:
            collector_checkers.append(checker_name)

        if ReturnValueCollector.checker_collect in checker_name:
            collector_checkers.append(checker_name)

    if not collector_checkers:
        LOG.debug('No available statistics collector checkers were found')
        return [], False

    for coll_check in collector_checkers:
        cmd.extend(['-Xclang', f'-analyzer-checker={coll_check}'])

    compile_lang = action.lang
    if not has_flag('-x', cmd):
        cmd.extend(['-x', compile_lang])

    if not has_flag('--target', cmd) and action.target != "":
        cmd.append(f"--target={action.target}")

    if not has_flag('-std', cmd) and not has_flag('--std', cmd):
        cmd.append(action.compiler_standard)

    cmd.extend(prepend_all('-isystem', action.compiler_includes))

    if source:
        cmd.append(source)
    return cmd, True



"""
System comments

Revision ID: 6cb6a3a41967
Revises:     3e91d0612422
Create Date: 2019-04-02 16:12:46.794131
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '6cb6a3a41967'
down_revision = '3e91d0612422'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('comments', sa.Column('kind',
                                        sa.Integer(),
                                        server_default="0",
                                        nullable=False),)


def downgrade():
    op.drop_column('comments', 'kind')



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Sometimes it might be useful to collect the source files constituting a
translation unit. This script does this action based on a compilation database
JSON file. The output of the script is a ZIP package with the collected
sources.
"""


import argparse
import collections
import fnmatch
import hashlib
import json
import logging
import os
import random
import re
import shlex
import string
import subprocess
import sys
import zipfile
from distutils.spawn import find_executable

from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Set, Tuple, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from mypy_extensions import TypedDict


LOG = logging.getLogger('tu_collector')

handler = logging.StreamHandler()
formatter = logging.Formatter('[%(levelname)s] - %(message)s')
handler.setFormatter(formatter)

LOG.setLevel(logging.INFO)
LOG.addHandler(handler)


class CompileAction(TypedDict):
    file: str
    command: str
    directory: str


CompilationDB = List[CompileAction]


def __random_string(length: int) -> str:
    """
    This function returns a random string of ASCII lowercase characters with
    the given length.
    """
    return ''.join(random.choice(string.ascii_lowercase)
                   for i in range(length))


def __get_toolchain_compiler(command: List[str]) -> Optional[str]:
    """
    Clang can be given a GCC toolchain so that the standard libs of that GCC
    are used. This function returns the path of the GCC toolchain compiler.
    """
    for cmp_opt in command:
        tcpath = re.match(r"^--gcc-toolchain=(?P<tcpath>.*)$", cmp_opt)
        if tcpath:
            is_cpp = '++' in command[0] or 'cpp' in command[0]
            return os.path.join(tcpath.group('tcpath'),
                                'bin',
                                'g++' if is_cpp else 'gcc')
    return None


def __determine_compiler(gcc_command: List[str]) -> str:
    """
    This function determines the compiler from the given compilation command.
    If the first part of the gcc_command is ccache invocation then the rest
    should be a complete compilation command.

    CCache may have three forms:
    1. ccache g++ main.cpp
    2. ccache main.cpp
    3. /usr/lib/ccache/gcc main.cpp
    In the first case this function drops "ccache" from gcc_command and returns
    the next compiler name.
    In the second case the compiler can be given by config files or an
    environment variable. Currently we don't handle this version, and in this
    case the compiler remanis "ccache" and the gcc_command is not changed.
    The two cases are distinguished by checking whether the second parameter is
    an executable or not.
    In the third case gcc is a symlink to ccache, but we can handle
    it as a normal compiler.

    gcc_command -- A split build action as a list which may or may not start
                   with ccache.

    !!!WARNING!!! This function must always return an element of gcc_command
    without modification (symlink resolve, absolute path conversion, etc.)
    otherwise an exception is thrown at the caller side.

    TODO: The second case could be handled if there was a way for querying the
    used compiler from ccache. This can be configured for ccache in config
    files or environment variables.
    """
    if gcc_command[0].endswith('ccache'):
        if find_executable(gcc_command[1]) is not None:
            return gcc_command[1]

    return gcc_command[0]


def __gather_dependencies(
    cmd: Union[str, List[str]],
    build_dir: str
) -> List[str]:
    """
    Returns a list of files which are contained in the translation unit built
    by the given build command.

    cmd -- The build command as a string or as a list that can be given to
           subprocess.Popen(). The first element is the executable
           compiler.
    build_dir -- The path of the working directory where the build command was
                 emitted.
    """

    def __eliminate_argument(
        arg_vect: List[str],
        opt_string: str,
        has_arg=False
    ) -> List[str]:
        """
        This call eliminates the parameters matching the given option string,
        along with its argument coming directly after the opt-string if any,
        from the command. The argument can possibly be separated from the flag.
        """
        while True:
            option_index = next(
                (i for i, c in enumerate(arg_vect)
                 if c.startswith(opt_string)), None)

            if option_index:
                separate = 1 if has_arg and \
                    len(arg_vect[option_index]) == len(opt_string) else 0
                arg_vect = arg_vect[0:option_index] + \
                    arg_vect[option_index + separate + 1:]
            else:
                break

        return arg_vect

    command = shlex.split(cmd) if isinstance(cmd, str) else cmd

    # gcc and clang can generate makefile-style dependency list.

    # If an output file is set, the dependency is not written to the
    # standard output but rather into the given file.
    # We need to first eliminate the output from the command.
    command = __eliminate_argument(command, '-o', True)
    command = __eliminate_argument(command, '--output', True)

    # This flag can be given a .specs file which contains the config options of
    # cc1, cc1plus, as, ld, etc. Sometimes this file is just a temporary during
    # the compilation. However, if the file doesn't exist, this flag fails the
    # compilation. Since this flag is not necessary for dependency generation,
    # we can skip it.
    command = __eliminate_argument(command, '-specs')

    # Remove potential dependency-file-generator options from the string
    # too. These arguments found in the logged build command would derail
    # us and generate dependencies, e.g. into the build directory used.
    command = __eliminate_argument(command, '-MM')
    command = __eliminate_argument(command, '-MF', True)
    command = __eliminate_argument(command, '-MP')
    command = __eliminate_argument(command, '-MT', True)
    command = __eliminate_argument(command, '-MQ', True)
    command = __eliminate_argument(command, '-MD')
    command = __eliminate_argument(command, '-MMD')

    # Clang contains some extra options.
    command = __eliminate_argument(command, '-MJ', True)
    command = __eliminate_argument(command, '-MV')

    # Build out custom invocation for dependency generation.
    compiler = __determine_compiler(command)
    command = [compiler, '-E', '-M', '-MT', '__dummy'] \
        + command[command.index(compiler) + 1:]

    # Remove empty arguments
    command = [i for i in command if i]

    # gcc does not have '--gcc-toolchain' argument it would fail if it is
    # kept there.
    # For clang it does not change the output, the include paths from
    # the gcc-toolchain are not added to the output.
    command = __eliminate_argument(command, '--gcc-toolchain')

    LOG.debug("Command: %s", ' '.join(command))

    try:
        output = subprocess.check_output(
            command,
            cwd=build_dir,
            encoding="utf-8",
            errors="replace")
        rc = 0
    except subprocess.CalledProcessError as ex:
        output, rc = ex.output, ex.returncode
    except OSError as oerr:
        output, rc = oerr.strerror, oerr.errno

    if rc != 0:
        raise IOError(output)

    # Parse 'Makefile' syntax dependency output.
    dependencies = output.replace('__dummy: ', '') \
        .replace('\\', '') \
        .replace('  ', '') \
        .replace(' ', '\n')

    # The dependency list already contains the source file's path.
    return [os.path.join(build_dir, dep) for dep in
            dependencies.splitlines() if dep != ""]


def __analyzer_action_hash(build_action: CompileAction) -> str:
    """
    This function returns a hash of a build action. This hash algorithm is
    duplicated based on the same algorithm in CodeChecker. It is important to
    keep these algorithms in sync!!! Basically the hash is computed from the
    build command except for -o flag. The exclusion is due to a special
    behavior of "make" utility. See the full description in CodeChecker:
    analyzer_action_str() contains the documentation of the issue.
    """
    source_file = os.path.normpath(
        os.path.join(build_action['directory'], build_action['file']))

    args = shlex.split(build_action['command'])
    indices = [idx for idx, v in enumerate(args) if v.startswith('-o')]

    for idx in reversed(indices):
        # Output can be given separate or joint:
        # -o a.out vs -oa.out
        # In the first case we delete its argument too.
        if args[idx] == '-o':
            del args[idx]
        del args[idx]

    build_info = source_file + '_' + ' '.join(args)

    return hashlib.md5(build_info.encode(errors='ignore')).hexdigest()


def __get_ctu_buildactions(
    build_action: CompileAction,
    compilation_db: CompilationDB,
    ctu_deps_dir: str
) -> Iterator[CompileAction]:
    """
    CodeChecker collets which source files were involved in CTU analysis. This
    function returns the build actions which describe the compilation of these
    files.
    build_action -- The build action object which is analyzed in CTU mode. Its
                    CTU dependencies will be returned.
    compilation_db -- A complete compilation database. This function returns
                      some of its elements: the ones which describe the
                      compilation of the files listed in the file under
                      ctu_deps_dir belonging to build_action.
    ctu_deps_dir -- A directory created by CodeChecker under the report
                    directory. The files under this folder must contain a hash
                    of a build action. See __analyzer_action_hash().
    """
    ctu_deps_file = next(
        (f for f in os.listdir(ctu_deps_dir)
         if __analyzer_action_hash(build_action) in f), None)

    if not ctu_deps_file:
        return iter(())

    with open(os.path.join(ctu_deps_dir, ctu_deps_file),
              encoding='utf-8', errors='ignore') as f:
        files = list(map(lambda x: x.strip(), f.readlines()))

    return filter(
        lambda ba: os.path.join(ba['directory'], ba['file']) in files,
        compilation_db)


def get_dependent_headers(
    command: Union[str, List[str]],
    build_dir_path: str,
    collect_toolchain=True
) -> Tuple[Set[str], str]:
    """
    Returns a pair of which the first component is a set of files building up
    the translation unit and the second component is an error message which is
    not empty in case some files may be missing from the set.

    command -- The build command as a string or as a list that can be given to
               subprocess.Popen(). The first element is the executable
               compiler.
    build_dir_path -- The path of the working directory where the build command
                      was emitted.
    collect_toolchain -- If the given command uses Clang and it is given a GCC
                         toolchain then the toolchain compiler's dependencies
                         are also collected in case this parameter is True.
    """

    LOG.debug("Generating dependent headers via compiler...")

    if isinstance(command, str):
        command = shlex.split(command)

    dependencies = set()
    error = ''

    try:
        dependencies |= set(__gather_dependencies(command, build_dir_path))
    except Exception as ex:
        LOG.error("Couldn't create dependencies: %s", str(ex))
        error += str(ex)

    toolchain_compiler = __get_toolchain_compiler(command)

    if collect_toolchain and toolchain_compiler:
        LOG.debug("Generating gcc-toolchain headers via toolchain "
                  "compiler...")
        try:
            # Change the original compiler to the compiler from the toolchain.
            command[0] = toolchain_compiler
            dependencies |= set(__gather_dependencies(command, build_dir_path))
        except Exception as ex:
            LOG.error("Couldn't create dependencies: %s", str(ex))
            error += str(ex)

    LOG.debug("Dependencies: %s", ', '.join(dependencies))
    return dependencies, error


def add_sources_to_zip(
    zip_file: Union[str, Path],
    files: Union[str, Iterable[str]]
):
    """
    This function adds source files to the ZIP file if those are not present
    yet. The files will be placed to the "sources-root" directory under the ZIP
    file.

    zip_file -- A ZIP file name as a string.
    files -- A file path or an iterable of file paths.
    """
    if isinstance(files, str):
        files = [files]

    with zipfile.ZipFile(zip_file, 'a') as archive:
        for f in files:
            archive_path = os.path.normpath(
                os.path.join('sources-root', f.lstrip(os.sep)))

            try:
                archive.getinfo(archive_path)
            except KeyError:
                archive.write(f, archive_path, zipfile.ZIP_DEFLATED)
            else:
                LOG.debug("'%s' is already in the ZIP file, won't add it "
                          "again!", f)


def zip_tu_files(
    zip_file: Union[str, Path],
    compilation_db: Union[str, CompilationDB],
    file_filter='*',
    write_mode='w',
    ctu_deps_dir: Optional[str] = None
):
    """
    Collects all files to a zip file which are required for the compilation of
    the translation units described by the given compilation database.
    If there are some files which couldn't be collected then a "no-sources"
    file will contain the error message of its reason.
    The function returns the set of files.

    zip_file -- A file name or a file object.
    compilation_database -- Either a path of the compilation database JSON file
                            or a list of the parsed JSON.
    file_filter -- A glob used as file name filter. The files of a TU will be
                   collected only if the "file" attribute of the build action
                   matches this glob.
    write_mode -- The file opening mode of the zip_file. In case of 'a' the new
                  files are appended to the existing zip file, in case of 'w'
                  the files are added to a clean zip file.
    ctu_deps_dir -- This directory contains a list of files. Each file belongs
                    to a translation unit and lists what other files are
                    involved during CTU analysis. These files and their
                    included headers will also be compressed in the .zip file.
                    File names in this directory must contain a hash of a the
                    build command. See __analyzer_action_hash() documentation.
                    When using this options, make sure to provide a compilation
                    database in the first argument which contains the build
                    commands of these files otherwise the files of this folder
                    can't be identified.
    """
    if isinstance(compilation_db, str):
        with open(compilation_db, encoding="utf-8", errors="ignore") as f:
            compilation_database = json.load(f)
    else:
        compilation_database = compilation_db

    no_sources = 'no-sources'
    tu_files: Set[str] = set()
    error_messages = ''

    filtered_compilation_database = list(filter(
        lambda action: fnmatch.fnmatch(action['file'], file_filter),
        compilation_database))

    if ctu_deps_dir:
        involved_ctu_actions: List[CompileAction] = []

        for action in filtered_compilation_database:
            involved_ctu_actions.extend(__get_ctu_buildactions(
                action, compilation_database, ctu_deps_dir))

        for action in involved_ctu_actions:
            if action not in filtered_compilation_database:
                filtered_compilation_database.append(action)

    for buildaction in filtered_compilation_database:
        files, err = get_dependent_headers(
            buildaction['command'],
            buildaction['directory'])

        tu_files |= files

        if err:
            error_messages += buildaction['file'] + '\n' \
                + '-' * len(buildaction['file']) + '\n' + err + '\n'

    if write_mode == 'a' and os.path.isfile(zip_file):
        with zipfile.ZipFile(zip_file) as archive:
            try:
                archive.getinfo(no_sources)
            except KeyError:
                pass
            else:
                no_sources = 'no-sources_' + __random_string(5)

    if error_messages:
        with zipfile.ZipFile(zip_file, write_mode) as archive:
            archive.writestr(no_sources, error_messages)
    elif write_mode == 'w':
        try:
            os.remove(zip_file)
        except OSError:
            pass

    add_sources_to_zip(zip_file, tu_files)

    with zipfile.ZipFile(zip_file, 'a') as archive:
        archive.writestr('compilation_database.json',
                         json.dumps(compilation_database, indent=2))


def get_dependent_sources(
    compilation_db: CompilationDB,
    header_path: Optional[str] = None
) -> Set[str]:
    """ Get dependencies for each files in each translation unit. """
    dependencies = collections.defaultdict(set)
    for build_action in compilation_db:
        files, _ = get_dependent_headers(
            build_action['command'],
            build_action['directory'])

        source_file = os.path.join(build_action['directory'],
                                   build_action['file'])
        for f in files:
            dependencies[f].add(source_file)

    pattern = None
    if header_path:
        norm_header_path = os.path.normpath(header_path.strip())
        pattern = re.compile(fnmatch.translate(norm_header_path))

    deps = set()
    for header, source_files in dependencies.items():
        if not pattern or pattern.match(header):
            deps.update(source_files)

    return deps


def main():
    # --- Handling of command line arguments --- #

    parser = argparse.ArgumentParser(
        description="""
This script can be used for multiple purposes:
- It can be used to collect all the source files constituting specific
translation units. The files are written to a ZIP file which will contain the
sources preserving the original directory hierarchy.
- It can be used to get source files which depend on a given header file.""",
        formatter_class=argparse.RawDescriptionHelpFormatter)

    log_args = parser.add_argument_group(
        "log arguments",
        """
Specify how the build information database should be obtained. You need to
specify either an already existing log file, or a build command which will be
used to generate a log file on the fly.""")
    log_args = log_args.add_mutually_exclusive_group(required=True)

    log_args.add_argument('-b', '--build', type=str, dest='command',
                          help="Execute and record a build command. Build "
                               "commands can be simple calls to 'g++' or "
                               "'clang++'.")
    log_args.add_argument('-l', '--logfile', type=str, dest='logfile',
                          help="Use an already existing JSON compilation "
                               "command database file specified at this path.")

    parser.add_argument('-f', '--filter', dest='filter',
                        type=str, required=False, default='*',
                        help="If '--zip' option is given this flag restricts "
                             "the collection on the build actions of which "
                             "the compiled source file matches this path. "
                             "If '--dependents' option is given this flag "
                             "specify a header file to get source file "
                             "dependencies for. E.g.: /path/to/*/files")
    parser.add_argument('--ctu-deps-dir', dest='ctu_deps_dir',
                        type=str, required=False,
                        help="When using 'CodeChecker analyze --ctu' command, "
                             "the results go to an output directory. This "
                             "folder contains a directory named "
                             "'ctu_connections' which stores the information "
                             "what TUs are involved during the analysis of "
                             "each source file. Once this directory is "
                             "provided, this tool also collects the sourced "
                             "of TUs involved by CTU analysis.")

    output_args = parser.add_argument_group(
        "output arguments",
        "Specify the output type.")

    output_args = output_args.add_mutually_exclusive_group(required=True)
    output_args.add_argument('-z', '--zip', dest='zip', type=str,
                             help="Output ZIP file.")

    output_args.add_argument('-d', '--dependents', dest='dependents',
                             action='store_true',
                             help="Use this flag to return a list of source "
                                  "files which depend on some header files "
                                  "specified by the --filter option. The "
                                  "result will not contain header files, even "
                                  "if those are dependents as well.")

    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        dest='verbose',
                        help="Enable debug level logging.")

    args = parser.parse_args()

    # --- Checking the existence of input files. --- #

    if 'verbose' in args and args.verbose:
        LOG.setLevel(logging.DEBUG)

    if args.logfile and not os.path.isfile(args.logfile):
        LOG.error("Compilation database file doesn't exist: %s", args.logfile)
        sys.exit(1)

    # --- Do the job. --- #

    if args.logfile:
        with open(args.logfile, encoding="utf-8", errors="ignore") as f:
            compilation_db = json.load(f)
    else:
        compilation_db = [{
            'file': '',
            'command': args.command,
            'directory': os.getcwd()}]

    if args.zip:
        if args.command and args.filter:
            LOG.warning('In case of using build command --filter has no '
                        'effect.')
        if args.command and args.ctu_deps_dir:
            LOG.warning('In case of using build command --ctu-deps-dir has no '
                        'effect.')

        zip_tu_files(args.zip, compilation_db, args.filter,
                     ctu_deps_dir=args.ctu_deps_dir)
        LOG.info("Done.")
    else:
        deps = get_dependent_sources(compilation_db, args.filter)

        if deps:
            print("\n".join(deps))
        else:
            LOG.info("No source file dependencies.")


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


from typing import List

from codechecker_report_converter.report import Report

from ...analyzer_result import AnalyzerResultBase
from .parser import Parser


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of Clang MemorySanitizer. """

    TOOL_NAME = 'msan'
    NAME = 'MemorySanitizer'
    URL = 'https://clang.llvm.org/docs/MemorySanitizer.html'

    def get_reports(self, file_path: str) -> List[Report]:
        """ Get reports from the given analyzer result. """
        return Parser().get_reports(file_path)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
from typing import List

from codechecker_report_converter.report import Report
from codechecker_report_converter.report.parser import sarif

from ..analyzer_result import AnalyzerResultBase


LOG = logging.getLogger('report-converter')


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of the GCC Static Analyzer. """

    TOOL_NAME = 'gcc'
    NAME = 'GNU Compiler Collection Static Analyzer'
    URL = 'https://gcc.gnu.org/wiki/StaticAnalyzer'

    def __init__(self):
        super(AnalyzerResult, self).__init__()

    def get_reports(self, result_file_path: str) -> List[Report]:
        """ Get reports from the given analyzer result file. """

        return sarif.Parser().get_reports(result_file_path)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Helper functions for Thrift api calls.
"""

from codechecker_api.codeCheckerDBAccess_v6 import codeCheckerDBAccess

from codechecker_client.thrift_call import ThriftClientCall
from .base import BaseClientHelper


class ThriftResultsHelper(BaseClientHelper):

    def __init__(self, protocol, host, port, uri, session_token=None,
                 get_new_token=None):
        """
        @param get_new_token: a function which can generate a new token.
        """
        super().__init__(protocol, host, port, uri, session_token,
                         get_new_token)

        self.client = codeCheckerDBAccess.Client(self.protocol)

    @ThriftClientCall
    def getRunData(self, run_name_filter, limit, offset, sort_mode):
        pass

    @ThriftClientCall
    def getRunHistory(self, run_ids, limit, offset, run_history_filter):
        pass

    @ThriftClientCall
    def getReportDetails(self, reportId):
        pass

    @ThriftClientCall
    def getSourceFileData(self, fileId, fileContent, encoding):
        pass

    @ThriftClientCall
    def getLinesInSourceFileContents(self, lines_in_files_requested, encoding):
        pass

    @ThriftClientCall
    def getDiffResultsHash(self, run_ids, report_hashes, diff_type,
                           skip_detection_statuses, tag_ids):
        pass

    @ThriftClientCall
    def getRunResultTypes(self, runId, reportFilters):
        pass

    @ThriftClientCall
    def removeRunResults(self, run_ids):
        pass

    @ThriftClientCall
    def removeRunReports(self, run_ids, report_filter, cmp_data):
        pass

    @ThriftClientCall
    def removeRun(self, run_id, run_filter):
        pass

    @ThriftClientCall
    def updateRunData(self, run_id, new_run_name):
        pass

    @ThriftClientCall
    def getSuppressedBugs(self, run_id):
        pass

    @ThriftClientCall
    def getNewResults(self, base_run_id, new_run_id, limit, offset, sortType,
                      reportFilters):
        pass

    @ThriftClientCall
    def getUnresolvedResults(self, base_run_id, new_run_id, limit, offset,
                             sortType, reportFilters):
        pass

    @ThriftClientCall
    def getResolvedResults(self, base_run_id, new_run_id, limit, offset,
                           sortType, reportFilters):
        pass

    @ThriftClientCall
    def changeReviewStatus(self, report_id, status, message):
        pass

    @ThriftClientCall
    def changeReviewStatusByHash(self, bug_hash, status, message):
        pass

    @ThriftClientCall
    def getReviewStatusRules(self, filter, sortMode, limit, offset):
        pass

    @ThriftClientCall
    def getRunResults(self, runIds, limit, offset, sortType, reportFilter,
                      cmpData, getDetails):
        pass

    @ThriftClientCall
    def getReportAnnotations(self, key):
        pass

    @ThriftClientCall
    def getRunResultCount(self, runIds, reportFilter, cmpData):
        pass

    @ThriftClientCall
    def getSeverityCounts(self, runIds, reportFilter, cmpData):
        pass

    @ThriftClientCall
    def getCheckerMsgCounts(self, runIds, reportFilter, cmpData, limit,
                            offset):
        pass

    @ThriftClientCall
    def getReviewStatusCounts(self, runIds, reportFilter, cmpData):
        pass

    @ThriftClientCall
    def getDetectionStatusCounts(self, runIds, reportFilter, cmpData):
        pass

    @ThriftClientCall
    def getFileCounts(self, runIds, reportFilter, cmpData, limit, offset):
        pass

    @ThriftClientCall
    def getCheckerCounts(self, base_run_ids, reportFilter, cmpData, limit,
                         offset):
        pass

    @ThriftClientCall
    def exportData(self, runId):
        pass

    @ThriftClientCall
    def importData(self, exportData):
        pass
    # SOURCE COMPONENT RELATED API CALLS

    @ThriftClientCall
    def addSourceComponent(self, name, value, description):
        pass

    @ThriftClientCall
    def getSourceComponents(self, component_filter):
        pass

    @ThriftClientCall
    def removeSourceComponent(self, name):
        pass

    # STORAGE RELATED API CALLS

    @ThriftClientCall
    def getMissingContentHashes(self, file_hashes):
        pass

    @ThriftClientCall
    def getMissingContentHashesForBlameInfo(self, file_hashes):
        pass

    @ThriftClientCall
    def massStoreRun(self, name, tag, version, zipdir, force,
                     trim_path_prefixes, description):
        pass

    @ThriftClientCall
    def allowsStoringAnalysisStatistics(self):
        pass

    @ThriftClientCall
    def getAnalysisStatisticsLimits(self):
        pass

    @ThriftClientCall
    def getAnalysisStatistics(self, run_id, run_history_id):
        pass

    @ThriftClientCall
    def storeAnalysisStatistics(self, run_name, zip_file):
        pass



"""
Store limit

Revision ID: 00099e8bc212
Revises:     7829789fc19c
Create Date: 2023-03-10 16:45:19.301602
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '00099e8bc212'
down_revision = '7829789fc19c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('products',
                  sa.Column('report_limit', sa.Integer(),
                            server_default='500000', nullable=False))


def downgrade():
    op.drop_column('products', 'report_limit')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
"""
from collections import defaultdict
# TODO distutils will be removed in python3.12
from distutils.version import StrictVersion
import os
import pickle
import shlex
import subprocess

from codechecker_common.logger import get_logger

from codechecker_analyzer import analyzer_context

from .. import analyzer_base
from ..flag import has_flag
from ..config_handler import CheckerState

from .config_handler import GccConfigHandler
from .result_handler import GccResultHandler, \
    actual_name_to_codechecker_name, codechecker_name_to_actual_name_disabled

LOG = get_logger('analyzer.gcc')


class Gcc(analyzer_base.SourceAnalyzer):
    """
    Constructs the Gcc analyzer commands.
    """

    ANALYZER_NAME = 'gcc'

    @classmethod
    def analyzer_binary(cls):
        return analyzer_context.get_context() \
            .analyzer_binaries[cls.ANALYZER_NAME]

    def add_checker_config(self, checker_cfg):
        # TODO
        pass

    @classmethod
    def get_analyzer_mentioned_files(self, output):
        """
        This is mostly used for CTU, which is absent in GCC.
        """
        pass

    def construct_analyzer_cmd(self, result_handler):
        """
        Construct analyzer command for gcc.
        """
        # TODO: This is not a try-catch block, like the other analyzers. Why
        # should it? Should the others be? When can list creating list to have
        # unforeseen exceptions where a general catch is justified?
        config = self.config_handler

        # We don't want GCC do start linking, but -fsyntax-only stops the
        # compilation process too early for proper diagnostics generation.
        analyzer_cmd = [Gcc.analyzer_binary(), '-fanalyzer', '-c',
                        '-o/dev/null']

        # Add extra arguments.
        analyzer_cmd.extend(config.analyzer_extra_arguments)

        analyzer_cmd.extend(self.buildaction.analyzer_options)

        analyzer_cmd.append('-fdiagnostics-format=sarif-stderr')

        for checker_name, value in config.checks().items():
            if value[0] == CheckerState.disabled:
                # TODO python3.9 removeprefix method would be nicer
                # than startswith and a hardcoded slicing
                analyzer_cmd.append(
                    codechecker_name_to_actual_name_disabled(checker_name))

        compile_lang = self.buildaction.lang
        if not has_flag('-x', analyzer_cmd):
            analyzer_cmd.extend(['-x', compile_lang])

        analyzer_cmd.append(self.source_file)

        LOG.debug_analyzer("Running analysis command "
                           f"'{shlex.join(analyzer_cmd)}'")

        return analyzer_cmd

    @classmethod
    def get_analyzer_checkers(self):
        """
        Return the list of the supported checkers.
        """
        command = [self.analyzer_binary(), "--help=warning"]
        checker_list = []

        try:
            output = subprocess.check_output(command)

            # Still contains the help message we need to remove.
            for entry in output.decode().split('\n'):
                warning_name, _, description = entry.strip().partition(' ')
                # GCC Static Analyzer names start with -Wanalyzer.
                if warning_name.startswith('-Wanalyzer'):
                    # Rename the checkers interally (similarly to how we
                    # support cppcheck)
                    renamed_checker_name = \
                        actual_name_to_codechecker_name(warning_name)
                    checker_list.append(
                        (renamed_checker_name, description.strip()))
            return checker_list
        except (subprocess.CalledProcessError) as e:
            LOG.error(e.stderr)
        except (OSError) as e:
            LOG.error(e.errno)
        return []

    @classmethod
    def get_analyzer_config(cls):
        """
        Config options for gcc.
        """
        # TODO
        return []

    @classmethod
    def get_checker_config(cls):
        """
        TODO add config options for gcc checkers.
        """
        # TODO
        return []

    def analyze(self, analyzer_cmd, res_handler, proc_callback=None):
        env = None

        original_env_file = os.environ.get(
            'CODECHECKER_ORIGINAL_BUILD_ENV')
        if original_env_file:
            with open(original_env_file, 'rb') as env_file:
                env = pickle.load(env_file, encoding='utf-8')

        return super().analyze(analyzer_cmd, res_handler, proc_callback, env)

    def post_analyze(self, result_handler: GccResultHandler):
        """
        Post process the reuslts after the analysis.
        Will copy the sarif files created by gcc into the root of the reports
        folder.
        Renames the source plist files to *.plist.bak because
        The report parsing of the Parse command is done recursively.

        """
        pass

    @classmethod
    def resolve_missing_binary(cls, configured_binary, env):
        """
        In case of the configured binary for the analyzer is not found in the
        PATH, this method is used to find a callable binary.
        """
        # TODO
        pass

    @classmethod
    def get_binary_version(self, environ, details=False) -> str:
        """
        Return the analyzer version.
        """
        # No need to LOG here, we will emit a warning later anyway.
        if not self.analyzer_binary():
            return None
        if details:
            version = [self.analyzer_binary(), '--version']
        else:
            version = [self.analyzer_binary(), '-dumpfullversion']
        try:
            output = subprocess.check_output(version,
                                             env=environ,
                                             encoding="utf-8",
                                             errors="ignore")
            return output.strip()
        except (subprocess.CalledProcessError, OSError) as oerr:
            LOG.warning("Failed to get analyzer version: %s",
                        ' '.join(version))
            LOG.warning(oerr)

        return None

    @classmethod
    def is_binary_version_incompatible(cls, environ):
        """
        Check the version compatibility of the given analyzer binary.
        """
        analyzer_version = cls.get_binary_version(environ)

        if analyzer_version is None:
            return "GCC binary is too old to support -dumpfullversion."

        # The analyzer version should be above 13.0.0 because the
        # '-fdiagnostics-format=sarif-file' argument was introduced in this
        # release.
        if analyzer_version >= StrictVersion("13.0.0"):
            return None

        return f"GCC binary found is too old at v{analyzer_version.strip()}; "\
               "minimum version is 13.0.0."

    def construct_result_handler(self, buildaction, report_output,
                                 skiplist_handler):
        """
        See base class for docs.
        """
        res_handler = GccResultHandler(buildaction, report_output,
                                       self.config_handler.report_hash)

        res_handler.skiplist_handler = skiplist_handler

        return res_handler

    @classmethod
    def construct_config_handler(cls, args):
        handler = GccConfigHandler()

        analyzer_config = defaultdict(list)

        if 'analyzer_config' in args and \
                isinstance(args.analyzer_config, list):
            for cfg in args.analyzer_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    analyzer_config[cfg.option].append(cfg.value)

        handler.analyzer_config = analyzer_config

        checkers = cls.get_analyzer_checkers()

        try:
            cmdline_checkers = args.ordered_checkers
        except AttributeError:
            LOG.debug_analyzer('No checkers were defined in '
                               'the command line for %s',
                               cls.ANALYZER_NAME)
            cmdline_checkers = []

        handler.initialize_checkers(
            checkers,
            cmdline_checkers,
            'enable_all' in args and args.enable_all)

        return handler



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
import logging
import sys
from typing import Optional, cast


class MigrationFormatter(logging.Formatter):
    """
    Truncates the filename to show only the revision that is being migrated
    in the log output.
    """
    def __init__(self, schema: str):
        super().__init__(fmt="[%(levelname)s][%(asctime)s] "
                             "{migration/%(schema)s} "
                             "[%(database)s] "
                             "- %(revision)s:%(lineno)d "
                             "- %(message)s",
                         datefmt="%Y-%m-%d %H:%M:%S")
        self.schema = schema
        self._database: Optional[str] = None

    @property
    def database(self):
        return self._database or "<UNKNOWN!>"

    @database.setter
    def database(self, db_name: str):
        self._database = db_name

    def format(self, record):
        record.database = self.database
        record.schema = self.schema
        record.revision = record.filename[:record.filename.find("_")]
        return super().format(record)


def setup_logger(schema: str):
    """
    Set up a logging system that should be used during schema migration.
    These outputs are not affected by the environment that executes a
    migration, e.g., by the running CodeChecker server!

    In migration scripts, use the built-in logging facilities instead of
    CodeChecker's wrapper, and ensure that the name of the logger created
    exactly matches ``migration/<some schema>``!
    """
    sys_logger = logging.getLogger("system")
    codechecker_loglvl = sys_logger.getEffectiveLevel()
    if codechecker_loglvl >= logging.INFO:
        # This might be 30 (WARNING) if the migration is run outside of
        # CodeChecker's context, e.g., in a downgrade.
        codechecker_loglvl = logging.INFO

    # Use the default logging class that came with Python for the migration,
    # temporarily turning away from potentially existing global changes.
    existing_logger_cls = logging.getLoggerClass()
    logging.setLoggerClass(logging.Logger)
    logger = logging.getLogger(f"migration/{schema}")
    logging.setLoggerClass(existing_logger_cls)

    if not logger.hasHandlers():
        fmt = MigrationFormatter(schema)
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        handler.setLevel(codechecker_loglvl)
        handler.setStream(sys.stdout)

        logger.setLevel(codechecker_loglvl)
        logger.addHandler(handler)
    else:
        handler = logger.handlers[0]
        fmt = handler.formatter

    return logger, handler, cast(MigrationFormatter, fmt)


def set_logger_database_name(schema: str, database: str):
    """
    Sets the logger's output for the current migration ``schema`` to indicate
    that the actions are performed on ``database``.
    """
    _, _, fmt = setup_logger(schema)
    fmt.database = database



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Build and log related functionality.
"""


import os
import pickle
import platform
import shlex
import subprocess
import sys
from uuid import uuid4

from codechecker_common.logger import get_logger

from .. import env
from . import host_check

LOG = get_logger('buildlogger')


def execute_buildcmd(command, silent=False, env=None, cwd=None):
    """
    Execute the the build command and continuously write
    the output from the process to the standard output.
    """
    proc = subprocess.Popen(
        command,
        bufsize=-1,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=cwd,
        shell=True,
        universal_newlines=True,
        encoding="utf-8",
        errors="ignore")

    while True:
        line = proc.stdout.readline()
        if not line and proc.poll() is not None:
            break
        if not silent:
            sys.stdout.write(line)

    return proc.returncode


def perform_build_command(logfile, command, keep_link, silent=False,
                          verbose=None):
    """
    Build the project and create a log file.
    """
    LOG.info("Starting build...")

    original_env = os.environ
    try:
        original_env_file = os.environ.get('CODECHECKER_ORIGINAL_BUILD_ENV')
        if original_env_file:
            LOG.debug_analyzer('Loading original build env from: %s',
                               original_env_file)

            with open(original_env_file, 'rb') as env_file:
                original_env = pickle.load(env_file, encoding='utf-8')

    except Exception as ex:
        LOG.warning(str(ex))
        LOG.warning('Failed to get saved original_env '
                    'using a current copy for logging.')
        original_env = os.environ.copy()

    # Run user's commands in shell, and preload ldlogger.
    # TODO: better platform detection.
    if host_check.check_ldlogger(os.environ) and platform.system() == 'Linux':
        LOG.info("Using CodeChecker ld-logger.")

        # Same as linux's touch.
        open(logfile, 'a', encoding="utf-8", errors="ignore").close()
        log_env = env.get_log_env(logfile, original_env)
        if 'CC_LOGGER_GCC_LIKE' not in log_env:
            log_env['CC_LOGGER_GCC_LIKE'] = 'gcc:g++:clang:clang++:cc:c++'
        if keep_link or ('CC_LOGGER_KEEP_LINK' in log_env and
                         log_env['CC_LOGGER_KEEP_LINK'] == 'true'):
            log_env['CC_LOGGER_KEEP_LINK'] = 'true'

        is_debug = verbose and verbose in ['debug', 'debug_analyzer']
        if not is_debug and 'CC_LOGGER_DEBUG_FILE' in log_env:
            del log_env['CC_LOGGER_DEBUG_FILE']
        elif is_debug and 'CC_LOGGER_DEBUG_FILE' not in log_env:
            if 'CC_LOGGER_DEBUG_FILE' in os.environ:
                log_file = os.environ['CC_LOGGER_DEBUG_FILE']
            else:
                log_file = os.path.join(os.path.dirname(logfile),
                                        'codechecker.logger.debug')

            if os.path.exists(log_file):
                os.remove(log_file)

            log_env['CC_LOGGER_DEBUG_FILE'] = log_file
    elif host_check.check_intercept(os.environ):
        LOG.info("Using intercept-build.")
        command = ' '.join(["intercept-build",
                            "--cdb", logfile,
                            "sh -c", shlex.quote(command)])
        log_env = original_env
        LOG.debug_analyzer(command)
    else:
        # Print a helpful diagnostic.
        if platform.system() == 'Linux':
            LOG.error("Both ldlogger and intercept-build are unavailable.\n"
                      "Try acquiring the compilation_commands.json in another "
                      "way.\n"
                      "Install ldlogger or intercept-build to proceed.")
            sys.exit(1)
        if platform.system() == 'Windows':
            LOG.error("This command is not supported on Windows. You can use "
                      "the following tools to generate a compilation "
                      "database: \n"
                      " - CMake (CMAKE_EXPORT_COMPILE_COMMANDS)\n"
                      " - compiledb (https://pypi.org/project/compiledb/)")
            sys.exit(1)
        if platform.system() == 'Darwin':
            LOG.error("Intercept-build is required to run CodeChecker in "
                      "OS X.")
            sys.exit(1)
        LOG.error("Unrecognized platform. Open a GitHub issue for further "
                  "guidance.")
        sys.exit(1)

    LOG.debug_analyzer(log_env)
    try:
        ret_code = execute_buildcmd(command, silent, log_env)

        if ret_code == 0:
            LOG.info("Build finished successfully.")
            LOG.debug_analyzer("The logfile is: %s", logfile)
        else:
            LOG.info("Build failed.")
            sys.exit(ret_code)

    except Exception as ex:
        LOG.error("Calling original build command failed.")
        LOG.error(str(ex))
        sys.exit(1)
    finally:
        debug_file = log_env.get('CC_LOGGER_DEBUG_FILE')
        if debug_file:
            LOG.info("The debug log file is: %s", debug_file)

            debug_logfile_lock = debug_file + '.lock'
            if os.path.exists(debug_logfile_lock):
                os.remove(debug_logfile_lock)

        # Removing flock lock file.
        logfile_lock = logfile + '.lock'
        if os.path.exists(logfile_lock):
            os.remove(logfile_lock)


def default_compilation_db(workspace_path, run_name):
    """
    Default compilation commands database file in the workspace.
    """
    workspace_path = os.path.abspath(workspace_path)
    uid = str(uuid4())[:10]  # 10 chars should be unique enough
    cmp_json_filename = 'compilation_commands_' + run_name + '_' \
                        + uid + '.json'
    compilation_commands = os.path.join(workspace_path, cmp_json_filename)
    return compilation_commands



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Check host machine for a compile command logger.
"""


import errno
import glob
import os
import subprocess

from codechecker_common.logger import get_logger

LOG = get_logger('system')


def check_intercept(env) -> bool:
    """
    Simple check if intercept (scan-build-py) is available.
    """
    intercept_cmd = ['intercept-build', '--help']
    try:
        res = subprocess.check_call(
            intercept_cmd,
            env=env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            encoding="utf-8",
            errors="ignore")

        if not res:
            return True
        else:
            LOG.debug('Failed to run: "%s"', ' '.join(intercept_cmd))
            return False
    except subprocess.CalledProcessError:
        LOG.debug('Failed to run: "%s", process returned non-zero exit code.',
                  ' '.join(intercept_cmd))
        return False
    except OSError as oerr:
        if oerr.errno == errno.ENOENT:
            # Not just intercept-build can be used for logging.
            # It is possible that another build logger is available.
            LOG.debug(oerr)
            LOG.debug('Failed to run: "%s"', ' '.join(intercept_cmd))
            return False


def check_ldlogger(env) -> bool:
    """
    Simple check if ldlogger is available.
    """

    data_files_dir_path = env.get('CC_DATA_FILES_DIR', '')
    lib_dir_path = env.get('CC_LIB_DIR', '')

    if not data_files_dir_path and not lib_dir_path:
        LOG.debug('Both CC_DATA_FILES_DIR and CC_LIB_DIR environment '
                  'variables as undefined or empty in '
                  'the logger environment.')
        return False

    logger_lib_dir_path = os.path.join(
        data_files_dir_path, 'ld_logger', 'lib', '**', 'ldlogger.so')
    if glob.glob(logger_lib_dir_path, recursive=True):
        return True

    alternative_logger_lib_dir_path = os.path.join(
        lib_dir_path, 'codechecker_analyzer', 'ld_logger', 'lib', '**',
        'ldlogger.so')
    return glob.glob(alternative_logger_lib_dir_path, recursive=True)



"""
Add bug event and point report index

Revision ID: 101a9cb747de
Revises:     dd9c97ead24
Create Date: 2018-02-15 15:30:59.966552
"""

from alembic import op


# Revision identifiers, used by Alembic.
revision = '101a9cb747de'
down_revision = 'dd9c97ead24'
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(op.f('ix_bug_path_events_report_id'), 'bug_path_events',
                    ['report_id'], unique=False)
    op.create_index(op.f('ix_bug_report_points_report_id'),
                    'bug_report_points', ['report_id'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_bug_path_events_report_id'),
                  table_name='bug_path_events')
    op.drop_index(op.f('ix_bug_report_points_report_id'),
                  table_name='bug_report_points')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
"""


import fnmatch
import re
import os

from codechecker_common.logger import get_logger

LOG = get_logger('system')


class SkipListHandler:
    """
    Skiplist file format:

    -/skip/all/source/in/directory*
    -/do/not/check/this.file
    +/dir/check.this.file
    -/dir/*
    """

    def __init__(self, skip_file_content=""):
        """
        Process the lines of the skip file.
        """
        self.__skip = []
        if not skip_file_content:
            skip_file_content = ""

        self.__skip_file_lines = [line.strip() for line
                                  in skip_file_content.splitlines()
                                  if line.strip() and
                                  not line.strip().startswith('#')]

        valid_lines = self.__check_line_format(self.__skip_file_lines)
        self.__gen_regex(valid_lines)

    def __gen_regex(self, skip_lines):
        """
        Generate a regular expression from the given skip lines
        and collect them for later match.

        The lines should be checked for validity before generating
        the regular expressions.
        """
        for skip_line in skip_lines:
            norm_skip_path = os.path.normpath(skip_line[1:].strip())
            rexpr = re.compile(
                fnmatch.translate(norm_skip_path + '*'))
            self.__skip.append((skip_line, rexpr))

    def __check_line_format(self, skip_lines):
        """
        Check if the skip line is given in a valid format.
        Returns the list of valid lines.
        """
        valid_lines = []
        for line in skip_lines:
            if len(line) < 2 or line[0] not in ['-', '+']:
                LOG.warning("Skipping malformed skipfile pattern: %s", line)
                continue

            valid_lines.append(line)

        return valid_lines

    @property
    def skip_file_lines(self):
        """
        List of the lines from the skip file without changes.
        """
        return self.__skip_file_lines

    def overwrite_skip_content(self, skip_lines):
        """
        Cleans out the already collected skip regular expressions
        and rebuilds the list from the given skip_lines.
        """
        self.__skip = []
        valid_lines = self.__check_line_format(skip_lines)
        self.__gen_regex(valid_lines)

    def should_skip(self, source):
        """
        Check if the given source should be skipped.
        Should the analyzer skip the given source file?
        """
        if not self.__skip:
            return False

        for line, rexpr in self.__skip:
            if rexpr.match(source):
                sign = line[0]
                return sign == '-'
        return False


class SkipListHandlers(list):
    def should_skip(self, file_path: str):
        """
        True if the given source should be skipped by any of the skip list
        handler.
        """
        return any(handler.should_skip(file_path) for handler in self)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Defines a subcommand for CodeChecker which prints version information.
"""

import argparse
import json

from codechecker_common import logger
from codechecker_common.output import USER_FORMATS


LOG = logger.get_logger('system')


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """

    return {
        'prog': 'CodeChecker version',
        'formatter_class': argparse.ArgumentDefaultsHelpFormatter,

        # Description is shown when the command's help is queried directly
        'description': "Print the version of CodeChecker package that is "
                       "being used.",

        # Help is shown when the "parent" CodeChecker command lists the
        # individual subcommands.
        'help': "Print the version of CodeChecker package that is being used."
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    parser.add_argument('-o', '--output',
                        dest='output_format',
                        required=False,
                        default='table',
                        choices=USER_FORMATS,
                        help="The format to use when printing the version.")

    logger.add_verbose_arguments(parser)
    parser.set_defaults(func=main)


def main(args):
    """
    Get and print the version information from the version config
    file and Thrift API definition.
    """
    logger.setup_logger(args.verbose if 'verbose' in args else None)

    output_format = args.output_format

    # Get analyzer version information if the module is available.
    analyzer_version = None
    try:
        from codechecker_analyzer.cmd.analyzer_version import Version
        analyzer_version = Version()
    except Exception:
        pass

    # Get web version information if the module is available.
    web_version = None
    try:
        from codechecker_web.cmd.web_version import Version
        web_version = Version()
    except Exception:
        pass

    # Print the version information.
    if output_format == "json":
        print(json.dumps({
            "analyzer":
                analyzer_version.to_dict() if analyzer_version else None,
            "web": web_version.to_dict() if web_version else None}))
    else:
        if analyzer_version:
            analyzer_version.print(output_format)
            print()

        if web_version:
            web_version.print(output_format)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Helper for thrift api calls.
"""


import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException

import codechecker_api_shared

from codechecker_common.logger import get_logger

LOG = get_logger('system')


def truncate_arg(arg, max_len=100):
    """ Truncate the given argument if the length is too large. """
    if isinstance(arg, str) and len(arg) > max_len:
        return arg[:max_len] + "..."

    return arg


def ThriftClientCall(function):
    """ Wrapper function for thrift client calls.
        - open and close transport,
        - log and handle errors
    """
    funcName = function.__name__

    def wrapper(self, *args, **kwargs):
        self.transport.open()
        func = getattr(self.client, funcName)
        try:
            try:
                return func(*args, **kwargs)
            except TApplicationException as ex:
                # If the session is expired we will try to reset the token and
                # call the API function again.
                if "Error code 401" not in ex.message:
                    raise ex

                # Generate a new token
                self._reset_token()

                return func(*args, **kwargs)
        except codechecker_api_shared.ttypes.RequestFailed as reqfailure:
            LOG.error('Calling API endpoint: %s', funcName)
            if reqfailure.errorCode ==\
                codechecker_api_shared.ttypes.ErrorCode.GENERAL and \
                    reqfailure.extraInfo and \
                    reqfailure.extraInfo[0] == "report_limit":
                # We handle this error in near the business logic.
                raise reqfailure
            else:
                if reqfailure.errorCode ==\
                        codechecker_api_shared.ttypes.ErrorCode.DATABASE:
                    LOG.error('Database error on server\n%s',
                              str(reqfailure.message))
                elif reqfailure.errorCode ==\
                        codechecker_api_shared.ttypes.ErrorCode.AUTH_DENIED:
                    LOG.error('Authentication denied\n %s',
                              str(reqfailure.message))
                elif reqfailure.errorCode ==\
                        codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED:
                    LOG.error('Unauthorized to access\n %s',
                              str(reqfailure.message))
                    LOG.error('Ask the product admin for additional access '
                              'rights.')
                elif reqfailure.errorCode ==\
                        codechecker_api_shared.ttypes.ErrorCode.API_MISMATCH:
                    LOG.error('Client/server API mismatch\n %s',
                              str(reqfailure.message))
                else:
                    LOG.error('API call error: %s\n%s',
                              funcName,
                              str(reqfailure)
                              )
                sys.exit(1)
        except TApplicationException as ex:
            LOG.error("Internal server error: %s", str(ex.message))
            sys.exit(1)
        except TProtocolException as ex:
            if ex.type == TProtocolException.UNKNOWN:
                LOG.error('Unknown thrift error')
            elif ex.type == TProtocolException.INVALID_DATA:
                LOG.error('Thrift invalid data error.')
            elif ex.type == TProtocolException.NEGATIVE_SIZE:
                LOG.error('Thrift negative size error.')
            elif ex.type == TProtocolException.SIZE_LIMIT:
                LOG.error('Thrift size limit error.')
            elif ex.type == TProtocolException.BAD_VERSION:
                LOG.error('Thrift bad version error.')
            LOG.error(funcName)

            # Do not print the argument list if it contains sensitive
            # information such as passwords.
            # Also it is possible that one of the argument is too large to log
            # the full content of it (for example the 'b64zip' parameter of the
            # 'massStoreRun' API function). For this reason we have to truncate
            # the arguments.
            if funcName != "performLogin":
                LOG.error([truncate_arg(arg) for arg in args])

            LOG.error(kwargs)
            LOG.exception("Request failed.")
            sys.exit(1)
        except OSError as oserr:
            LOG.error("Connection failed.")
            LOG.error(oserr.strerror)
            LOG.error("Check if your CodeChecker server is running.")
            sys.exit(1)
        finally:
            self.transport.close()

    return wrapper



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
'CodeChecker log' executes a build action and registers a compilation database
for the given build, using an external tool such as scan-build-py or ld-logger.

This module contains the basic definitions for how 'CodeChecker log' is to be
invoked and ran.
"""


import argparse
import os

from codechecker_analyzer.buildlog import build_manager
from codechecker_analyzer.buildlog.host_check import check_intercept
from codechecker_analyzer.buildlog.host_check import check_ldlogger

from codechecker_common import arg, logger


epilog_env_var = f"""
  CC_LOGGER_ABS_PATH       If the environment variable is defined, all relative
                           paths in the compilation commands after '-I,
                           -idirafter, -imultilib, -iquote, -isysroot -isystem,
                           -iwithprefix, -iwithprefixbefore, -sysroot,
                           --sysroot' will be converted to absolute PATH when
                           written into the compilation database.
  CC_LOGGER_DEBUG_FILE     Output file to print log messages. By default if we
                           run the log command in debug mode it will generate
                           a 'codechecker.logger.debug' file beside the log
                           file.
  CC_LOGGER_DEF_DIRS       If the environment variable is defined, the logger
                           will extend the compiler argument list in the
                           compilation database with the pre-configured include
                           paths of the logged compiler.
  CC_LOGGER_GCC_LIKE       Set to to a colon separated list to change which
                           compilers should be logged. For example (default):
                           export CC_LOGGER_GCC_LIKE="gcc:g++:clang:clang++:
                           cc:c++". The logger will match any compilers with
                           'gcc', 'g++', 'clang', 'clang++', 'cc' and 'c++' in
                           their filenames.
  CC_LOGGER_KEEP_LINK      If its value is not 'true' then object files will be
                           removed from the build action. For example in case
                           of this build command: 'gcc main.c object1.o
                           object2.so' the 'object1.o' and 'object2.so' will be
                           removed and only 'gcc main.c' will be captured. If
                           only object files are provided to the compiler then
                           the complete build action will be thrown away. This
                           means that build actions which only perform linking
                           will not be captured. We consider a file as object
                           file if its extension is '.o', '.so' or '.a'.
"""


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """

    # Prefer using ldlogger over intercept-build.
    is_ldlogger = check_ldlogger(os.environ)
    is_intercept = False if is_ldlogger else check_intercept(os.environ)
    ldlogger_settings = """
ld-logger can be fine-tuned with some environment variables. For details see
the following documentation:
https://github.com/Ericsson/codechecker/blob/master/analyzer/tools/
build-logger/README.md#usage""" if not is_ldlogger else ""

    return {
        'prog': 'CodeChecker log',
        'formatter_class': arg.RawDescriptionDefaultHelpFormatter,

        # Description is shown when the command's help is queried directly
        'description': """
Runs the given build command and records the executed compilation steps. These
steps are written to the output file in a JSON format.

Available build logger tool that will be used is '{0}'.{1}
""".format('intercept-build' if is_intercept else 'ld-logger',
           ldlogger_settings),

        'epilog': f"""
Environment variables
------------------------------------------------
{epilog_env_var}
""",

        # Help is shown when the "parent" CodeChecker command lists the
        # individual subcommands.
        'help': "Run a build command and collect the executed compilation "
                "commands, storing them in a JSON file."
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    parser.add_argument('-o', '--output',
                        type=str,
                        dest="logfile",
                        default=argparse.SUPPRESS,
                        required=True,
                        help="Path of the file to write the collected "
                             "compilation commands to. If the file already "
                             "exists, it will be overwritten.")

    parser.add_argument('-b', '--build',
                        type=str,
                        dest="command",
                        default=argparse.SUPPRESS,
                        required=True,
                        help="The build command to execute. Build commands "
                             "can be simple calls to 'g++' or 'clang++' or "
                             "'make', but a more complex command, or the call "
                             "of a custom script file is also supported.")

    parser.add_argument('-k', '--keep-link',
                        dest="keep_link",
                        default=argparse.SUPPRESS,
                        action="store_true",
                        help="If this flag is given then the output log file "
                             "will contain the linking build actions too.")

    parser.add_argument('-q', '--quiet',
                        dest="quiet",
                        action='store_true',
                        default=argparse.SUPPRESS,
                        required=False,
                        help="Do not print the output of the build tool into "
                             "the output of this command.")

    parser.add_argument('--verbose',
                        type=str,
                        dest='verbose',
                        choices=logger.CMDLINE_LOG_LEVELS,
                        default=argparse.SUPPRESS,
                        help="Set verbosity level. If the value is 'debug' or "
                             "'debug_analyzer' it will create a "
                             "'codechecker.logger.debug' debug log file "
                             "beside the given output file. It will contain "
                             "debug information of compilation database "
                             "generation. You can override the location of "
                             "this file if you set the 'CC_LOGGER_DEBUG_FILE' "
                             "environment variable to a different file path.")

    parser.set_defaults(func=main)


def main(args):
    """
    Generates a build log by running the original build command.
    No analysis is done.
    """
    logger.setup_logger(args.verbose if 'verbose' in args else None)

    args.logfile = os.path.realpath(args.logfile)

    # It is possible that the log file will not be created or it will be empty
    # for example when the build command is an empty string or when there is no
    # compiler command to log. For this reason we will create this log file if
    # it does not exist and we will insert an empty array to it, so it will be
    # a valid json file.
    with open(args.logfile, 'w',
              encoding="utf-8", errors="ignore") as logfile:
        logfile.write("[\n]")

    verbose = args.verbose if 'verbose' in args else None

    build_manager.perform_build_command(args.logfile,
                                        args.command,
                                        'keep_link' in args,
                                        silent='quiet' in args,
                                        verbose=verbose)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os
import re

from typing import Iterator, List, Optional, Tuple

from codechecker_report_converter.report import BugPathEvent, File, \
    get_or_create_file, Report

from ..parser import get_next, BaseParser

LOG = logging.getLogger('report-converter')


class SANParser(BaseParser):
    """ Parser for Clang UndefinedBehaviourSanitizer console outputs.

    Example output
        /a/b/main.cpp:13:10: runtime error: load of value 7...
    """

    checker_name: str = ""
    main_event_index = -1
    line_re = re.compile(r'')

    def __init__(self):
        super(SANParser, self).__init__()

        # Regex for parsing stack trace line.
        # It has the following format:
        #     #1 0x42a51d in main /dummy/main.cpp:24:2
        self.stack_trace_re = re.compile(r'^\s+#\d+')

        self.file_re = re.compile(
            r'(?P<path>[\S]+?):(?P<line>\d+)(:(?P<column>\d+))?')

    def parse_sanitizer_message(
        self,
        it: Iterator[str],
        line: str
    ) -> Tuple[Optional[Report], str]:
        """ Parses ThreadSanitizer output message. """
        match = self.line_re.match(line)
        if not match:
            return None, line

        if getattr(self, 'skip_line_after_match', True):
            line = get_next(it)
        stack_traces, events, line = self.parse_stack_trace(it, line)

        if not events:
            return None, line

        main_event = events[self.main_event_index]

        report = self.create_report(
            events, main_event.file, main_event.line, main_event.column,
            match.group('message').strip(), stack_traces)

        return report, line

    def _parse_line(
        self,
        it: Iterator[str],
        line: str
    ) -> Tuple[List[Report], str]:
        """ Parse the given line. """
        report, next_line = self.parse_sanitizer_message(it, line)
        if report:
            return [report], next_line

        return [], next(it)

    def parse_stack_trace_line(self, line: str) -> Optional[BugPathEvent]:
        """ Parse the given stack trace line.

        Return an event if the file in the stack trace line exists otherwise
        it returns None.
        """
        file_match = self.file_re.search(line)
        if not file_match:
            return None

        file_path = file_match.group('path')
        if file_path and os.path.exists(file_path):
            col = file_match.group('column')
            return BugPathEvent(
                line.rstrip(),
                get_or_create_file(
                    os.path.abspath(file_path), self._file_cache),
                int(file_match.group('line')),
                int(col) if col else 0)

        return None

    def create_report(
        self,
        events: List[BugPathEvent],
        file: File,
        line: int,
        column: int,
        message: str,
        stack_traces: List[str],
        checker_name: Optional[str] = None
    ) -> Report:
        """
        Create a report for the sanitizer output with the given data.
        checker_name: if None then the sanitizer name will be applied.
        """
        # The original message should be the last part of the path. This is
        # displayed by quick check, and this is the main event displayed by
        # the web interface.
        events.append(BugPathEvent(message, file, line, column))

        notes = None
        if stack_traces:
            notes = [BugPathEvent(''.join(stack_traces), file, line, column)]

        # Dynamic analyzers usually generate checker messages containing
        # memory addresses. Since checker messages are included in the report
        # hash, the subsequent executions of the analysis produces different
        # reports even if they refer the same problem. For this reason the
        # memory addresses are cut out from "static_message" for sake of
        # stabile hash generation.
        return Report(
            file, line, column, message, checker_name or self.checker_name,
            bug_path_events=events,
            notes=notes,
            static_message=re.sub(r'0x[0-9a-fA-F]+', '', message))

    def parse_stack_trace(
        self,
        it: Iterator[str],
        line: str
    ) -> Tuple[List[str], List[BugPathEvent], str]:
        """ Iterate over lines and parse stack traces. """
        events: List[BugPathEvent] = []
        stack_traces: List[str] = []

        while line.strip():
            event = self.parse_stack_trace_line(line)
            if event:
                events.append(event)

            stack_traces.append(line)
            line = get_next(it)

        events.reverse()

        return stack_traces, events, line



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Temporary directory module.
"""


import datetime
import hashlib
import os


from codechecker_common.logger import get_logger

LOG = get_logger('system')


def get_tmp_dir_hash():
    """Generate a hash based on the current time and process id."""

    pid = os.getpid()
    time = datetime.datetime.now()

    data = str(pid) + str(time)

    dir_hash = hashlib.md5()
    dir_hash.update(data.encode("utf-8"))

    LOG.debug('The generated temporary directory hash is %s.',
              dir_hash.hexdigest())

    return dir_hash.hexdigest()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


import logging
import re

from ..parser import SANParser


LOG = logging.getLogger('report-converter')


class Parser(SANParser):
    """ Parser for Clang MemorySanitizer console outputs. """

    checker_name = "MemorySanitizer"

    # Regex for parsing MemorySanitizer output message.
    line_re = re.compile(
        # Error code
        r'==(?P<code>\d+)==(ERROR|WARNING): MemorySanitizer: '
        # Checker message.
        r'(?P<message>[\S \t]+)')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os
import re

from typing import Iterable, Optional, Tuple

from codechecker_report_converter.report import get_or_create_file, Report

from ...parser import get_next
from ..parser import SANParser


LOG = logging.getLogger('report-converter')


class Parser(SANParser):
    """ Parser for Clang UndefinedBehaviourSanitizer console outputs.

    Example output
        /a/b/main.cpp:13:10: runtime error: load of value 7...
    """

    checker_name = "UndefinedBehaviorSanitizer"

    # Regex for parsing UndefinedBehaviorSanitizer output message.
    line_re = re.compile(
        # File path followed by a ':'.
        r'^(?P<path>[\S ]+?):'
        # Line number followed by a ':'.
        r'(?P<line>\d+?):'
        # Column number followed by a ':' and a space.
        r'(?P<column>\d+?): runtime error: '
        # Checker message.
        r'(?P<message>[\S \t]+)')

    # This list of diagnostic messages is extracted from the LLVM project:
    #   https://github.com/llvm/llvm-project
    # /blob/main/compiler-rt/lib/ubsan/ubsan_handlers.cpp
    # /blob/main/compiler-rt/lib/ubsan/ubsan_handlers_cxx.cpp
    # /blob/main/compiler-rt/lib/ubsan/ubsan_checks.inc
    checks = {
        "alignment": [
            re.compile(
                ".* misaligned address .* for type .*, which requires .* byte "
                "alignment"),
            re.compile(
                "assumption of .* byte alignment for pointer of type .* "
                "failed"),
            re.compile(
                "assumption of .* byte alignment (with offset of .* byte) for "
                "pointer of type .* failed")],
        "bool": [
            re.compile(
                "load of value .*, which is not a valid value for type "
                ".*bool.*")],
        "builtin": [
            re.compile("passing zero to .*, which is not a valid argument")],
        "bounds": [
            re.compile("index .* out of bounds for type .*")],
        "enum": [
            re.compile(
                "load of value .*, which is not a valid value for type .*")],
        "float-cast-overflow": [
            re.compile(
                ".* is outside the range of representable values of type .*")],
        "integer-divide-by-zero-or-float-divide-by-zero": [
            re.compile("division by zero")],
        "implicit-signed-integer-truncation": [
            re.compile(
                "implicit conversion from type .* of value .* (.*-bit, "
                "signed) to type .* changed the value to .* (.*-bit, "
                "signed)")],
        "implicit-unsigned-integer-truncation": [
            re.compile(
                "implicit conversion from type .* of value .* (.*-bit, "
                "unsigned) to type .* changed the value to .* (.*-bit, "
                "unsigned)")],
        "implicit-integer-sign-change": [
            re.compile(
                "implicit conversion from type .* of value .* (.*-bit, "
                ".*signed) to type .* changed the value to .* (.*-bit, "
                ".*signed)")],
        "nonnull-attribute-or-nullability-arg": [
            re.compile(
                "null pointer passed as argument .*, which is declared to "
                "never be null")],
        "null-or-nullability-assign": [
            re.compile(".* null pointer of type .*")],
        "nullability-return-or-returns-nonnull-attribute": [
            re.compile(
                "null pointer returned from function declared to never return "
                "null")],
        "objc-cast": [
            re.compile(
                "invalid ObjC cast, object is a '.*', but expected a .*")],
        "object-size": [
            re.compile(
                ".* address .* with insufficient space for an object of type "
                ".*")],
        "pointer-overflow": [
            re.compile("applying zero offset to null pointer"),
            re.compile("applying non-zero offset .* to null pointer"),
            re.compile(
                "applying non-zero offset to non-null pointer .* produced "
                "null pointer"),
            re.compile("addition of unsigned offset to .* overflowed to .*"),
            re.compile(
                "subtraction of unsigned offset from .* overflowed to .*"),
            re.compile(
                "pointer index expression with base .* overflowed to .*")],
        "return": [
            re.compile(
                "execution reached the end of a value-returning function "
                "without returning a value")],
        "shift": [
            re.compile("shift exponent .* is negative"),
            re.compile("shift exponent .* is too large for .*-bit type .*"),
            re.compile("left shift of negative value .*")],
        "signed-integer-overflow": [
            re.compile(
                "signed integer overflow: .* .* .* cannot be represented in "
                "type .*"),
            re.compile(
                "negation of .* cannot be represented in type .*; cast to an "
                "unsigned type to negate this value to itself"),
            re.compile(
                "division of .* by -1 cannot be represented in type .*")],
        "unreachable": [
            re.compile("execution reached an unreachable program point")],
        "unsigned-integer-overflow": [
            re.compile(
                "unsigned integer overflow: .* .* .* cannot be represented "
                "in type .*"),
            re.compile("negation of .* cannot be represented in type .*"),
            re.compile(
                "left shift of .* by .* places cannot be represented in type "
                ".*")],
        "vla-bound": [
            re.compile(
                "variable length array bound evaluates to non-positive value "
                ".*")],
        "vptr": [
            re.compile(
                ".* address .* which does not point to an object of type .*")]
    }

    def parse_stack_trace(self, it, line):
        """ Iterate over lines and parse stack traces. """
        events = []
        stack_traces = []

        while self.stack_trace_re.match(line):
            event = self.parse_stack_trace_line(line)
            if event:
                events.append(event)

            stack_traces.append(line)
            line = get_next(it)

        events.reverse()

        return stack_traces, events, line

    def deduce_checker_name(self, message: str) -> str:
        for check, patterns in self.checks.items():
            if any(pattern.search(message) for pattern in patterns):
                return f"{self.checker_name}.{check}"
        return self.checker_name

    def parse_sanitizer_message(
        self,
        it: Iterable[str],
        line: str
    ) -> Tuple[Optional[Report], str]:
        """ Parses UndefinedBehaviorSanitizer output message. """
        match = self.line_re.match(line)
        if not match:
            return None, line

        report_file = get_or_create_file(
            os.path.abspath(match.group('path')), self._file_cache)
        report_line = int(match.group('line'))
        report_col = int(match.group('column'))

        line = get_next(it)
        stack_traces, events, line = self.parse_stack_trace(it, line)

        message = match.group('message').strip()
        checker_name = self.deduce_checker_name(message)

        report = self.create_report(
            events, report_file, report_line, report_col,
            message, stack_traces, checker_name)

        return report, line



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Contains housekeeping routines that are used to remove expired, obsolete,
or dangling records from the database.
"""
from datetime import datetime, timedelta
from typing import Dict

import sqlalchemy

from codechecker_api.codeCheckerDBAccess_v6.ttypes import Severity

from codechecker_common import util
from codechecker_common.logger import get_logger

from .database import DBSession
from .run_db_model import \
    AnalysisInfo, \
    BugPathEvent, BugReportPoint, \
    Comment, Checker, \
    File, FileContent, \
    Report, ReportAnalysisInfo, RunHistoryAnalysisInfo, RunLock

LOG = get_logger('server')
RUN_LOCK_TIMEOUT_IN_DATABASE = 30 * 60  # 30 minutes.
SQLITE_LIMIT_COMPOUND_SELECT = 500


def remove_expired_data(product):
    """Remove information that has timed out from the database."""
    remove_expired_run_locks(product)


def remove_unused_data(product):
    """Remove dangling data (files, comments, etc.) from the database."""
    remove_unused_files(product)
    remove_unused_comments(product)
    remove_unused_analysis_info(product)


def update_contextual_data(product, context):
    """
    Updates information in the database that comes from potentially changing
    contextual configuration of the server package.
    """
    upgrade_severity_levels(product, context.checker_labels)


def remove_expired_run_locks(product):
    with DBSession(product.session_factory) as session:
        LOG.debug("[%s] Garbage collection of expired run locks started...",
                  product.endpoint)
        try:
            locks_expired_at = datetime.now() - timedelta(
                seconds=RUN_LOCK_TIMEOUT_IN_DATABASE)

            count = session.query(RunLock) \
                .filter(RunLock.locked_at < locks_expired_at) \
                .delete(synchronize_session=False)
            if count:
                LOG.debug("%d expired run locks deleted.", count)

            session.commit()

            LOG.debug("[%s] Garbage collection of expired run locks "
                      "finished.", product.endpoint)
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("[%s] Failed to remove expired run locks: %s",
                      product.endpoint, str(ex))


def remove_unused_files(product):
    # File deletion is a relatively slow operation due to database cascades.
    # Removing files in big chunks prevents reaching a potential database
    # statement timeout. This hard-coded value is a safe choice according to
    # some measurements. Maybe this could be a command-line parameter. But in
    # the long terms we are planning to reduce cascade deletes by redesigning
    # bug_path_events and bug_report_points tables.
    CHUNK_SIZE = 500_000
    with DBSession(product.session_factory) as session:
        LOG.debug("[%s] Garbage collection of dangling files started...",
                  product.endpoint)
        try:
            bpe_files = session.query(BugPathEvent.file_id) \
                .group_by(BugPathEvent.file_id) \
                .subquery()
            brp_files = session.query(BugReportPoint.file_id) \
                .group_by(BugReportPoint.file_id) \
                .subquery()

            files_to_delete = session.query(File.id) \
                .filter(File.id.notin_(bpe_files), File.id.notin_(brp_files))
            files_to_delete = map(lambda x: x[0], files_to_delete)

            total_count = 0
            for chunk in util.chunks(iter(files_to_delete), CHUNK_SIZE):
                q = session.query(File) \
                    .filter(File.id.in_(chunk))
                count = q.delete(synchronize_session=False)
                if count:
                    total_count += count

            if total_count:
                LOG.debug("%d dangling files deleted.", total_count)

            files = session.query(File.content_hash) \
                .group_by(File.content_hash) \
                .subquery()

            session.query(FileContent) \
                .filter(FileContent.content_hash.notin_(files)) \
                .delete(synchronize_session=False)

            session.commit()

            LOG.debug("[%s] Garbage collection of dangling files finished.",
                      product.endpoint)
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("[%s] Failed to remove unused files: %s",
                      product.endpoint, str(ex))


def remove_unused_comments(product):
    with DBSession(product.session_factory) as session:
        LOG.debug("[%s] Garbage collection of dangling comments started...",
                  product.endpoint)
        try:
            sub = session.query(Comment.id) \
                .join(Report,
                      Comment.bug_hash == Report.bug_id,
                      isouter=True) \
                .filter(Report.id.is_(None)) \
                .subquery()
            count = session.query(Comment) \
                .filter(Comment.id.in_(sub)) \
                .delete(synchronize_session=False)
            if count:
                LOG.debug("%d dangling comments deleted.", count)

            session.commit()

            LOG.debug("[%s] Garbage collection of dangling comments "
                      "finished.", product.endpoint)
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("[%s] Failed to remove dangling comments: %s",
                      product.endpoint, str(ex))


def remove_unused_analysis_info(product):
    with DBSession(product.session_factory) as session:
        LOG.debug("[%s] Garbage collection of dangling analysis info "
                  "started...", product.endpoint)
        try:
            # Disable foreign key constraints to avoid slow delete in Postgres
            if session.bind.dialect.name == "postgresql":
                rh_ai_foreign_keys = get_foreign_keys(
                    session,
                    RunHistoryAnalysisInfo.name,
                    AnalysisInfo.__tablename__
                )
                drop_foreign_keys(session,
                                  RunHistoryAnalysisInfo.name,
                                  rh_ai_foreign_keys)

                rep_ai_foreign_keys = get_foreign_keys(
                    session,
                    ReportAnalysisInfo.name,
                    AnalysisInfo.__tablename__
                )
                drop_foreign_keys(session,
                                  ReportAnalysisInfo.name,
                                  rep_ai_foreign_keys)

            to_delete = session.query(AnalysisInfo.id) \
                .join(
                    RunHistoryAnalysisInfo,
                    RunHistoryAnalysisInfo.c.analysis_info_id ==
                    AnalysisInfo.id,
                    isouter=True) \
                .join(
                    ReportAnalysisInfo,
                    ReportAnalysisInfo.c.analysis_info_id == AnalysisInfo.id,
                    isouter=True) \
                .filter(
                    RunHistoryAnalysisInfo.c.analysis_info_id.is_(None),
                    ReportAnalysisInfo.c.analysis_info_id.is_(None)) \
                .subquery()

            count = session.query(AnalysisInfo) \
                .filter(AnalysisInfo.id.in_(to_delete)) \
                .delete(synchronize_session=False)

            if count:
                LOG.debug("[%s] %d dangling analysis info deleted.",
                          product.endpoint, count)

            session.commit()

            LOG.debug("[%s] Garbage collection of dangling analysis info "
                      "finished.", product.endpoint)
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("[%s] Failed to remove dangling analysis info: %s",
                      product.endpoint, str(ex))
        finally:
            # Re-enable foreign key constraints
            if session.bind.dialect.name == "postgresql":
                add_foreign_keys(session,
                                 RunHistoryAnalysisInfo.name,
                                 rh_ai_foreign_keys)
                add_foreign_keys(session,
                                 ReportAnalysisInfo.name,
                                 rep_ai_foreign_keys)


def upgrade_severity_levels(product, checker_labels):
    """
    Updates the potentially changed severities to reflect the data in the
    current label configuration files.
    """
    with DBSession(product.session_factory) as session:
        LOG.debug("[%s] Upgrading severity levels started...",
                  product.endpoint)
        try:
            count = 0
            for analyzer in sorted(checker_labels.get_analyzers()):
                checkers_for_analyzer_in_database = session \
                    .query(Checker.id,
                           Checker.checker_name,
                           Checker.severity) \
                    .filter(Checker.analyzer_name == analyzer) \
                    .all()
                for checker_row in checkers_for_analyzer_in_database:
                    checker: str = checker_row.checker_name
                    old_severity_db: int = checker_row.severity
                    try:
                        old_severity: str = \
                            Severity._VALUES_TO_NAMES[old_severity_db]
                    except KeyError:
                        LOG.error("[%s] Checker '%s/%s' contains invalid "
                                  "severity %d, considering as if "
                                  "'UNSPECIFIED' (0)!",
                                  product.endpoint,
                                  analyzer, checker,
                                  old_severity_db)
                        old_severity_db, old_severity = 0, "UNSPECIFIED"
                    new_severity: str = \
                        checker_labels.severity(checker, analyzer)

                    if old_severity == new_severity:
                        continue

                    if new_severity == "UNSPECIFIED":
                        # No exact match for the checker's name in the
                        # label config for the analyzer. This can mean that
                        # the records are older than a change in the checker
                        # naming scheme (e.g., cppchecker results pre-2021).
                        LOG.warning("[%s] Checker '%s/%s' (database severity: "
                                    "'%s' (%d)) does not have a corresponding "
                                    "entry in the label config file.",
                                    product.endpoint,
                                    analyzer, checker,
                                    old_severity, old_severity_db)

                        new_sev_attempts: Dict[str, str] = {
                            chk_name: severity
                            for chk_name, severity in
                            ((name_attempt,
                              checker_labels.severity(name_attempt, analyzer))
                             for name_attempt in [
                                 "%s.%s" % (analyzer, checker),
                                 "%s-%s" % (analyzer, checker),
                                 "%s/%s" % (analyzer, checker)
                             ])
                            if severity != "UNSPECIFIED"
                        }

                        if len(new_sev_attempts) == 0:
                            LOG.debug("[%s] %s/%s: Keeping the old severity "
                                      "intact...",
                                      product.endpoint, analyzer, checker)
                            continue
                        if len(new_sev_attempts) >= 2 and \
                                len(set(new_sev_attempts.values())) >= 2:
                            LOG.error("[%s] %s/%s: Multiple similar checkers "
                                      "WITH CONFLICTING SEVERITIES were "
                                      "found instead: %s",
                                      product.endpoint,
                                      analyzer, checker,
                                      str(list(new_sev_attempts.items())))
                            LOG.debug("[%s] %s/%s: Keeping the old severity "
                                      "intact...",
                                      product.endpoint, analyzer, checker)
                            continue
                        if len(set(new_sev_attempts.values())) == 1:
                            attempted_name, new_severity = \
                                next(iter(sorted(new_sev_attempts.items())))

                            LOG.info("[%s] %s/%s: Found similar checker "
                                     "'%s/%s' (severity: '%s'), using this "
                                     "for the upgrade.",
                                     product.endpoint,
                                     analyzer, checker,
                                     analyzer, attempted_name,
                                     new_severity)
                            if old_severity == new_severity:
                                continue

                    new_severity_db: int = \
                        Severity._NAMES_TO_VALUES[new_severity]

                    LOG.info("[%s] Upgrading the severity of checker "
                             "'%s/%s' from '%s' (%d) to '%s' (%d).",
                             product.endpoint,
                             analyzer, checker,
                             old_severity, old_severity_db,
                             new_severity, new_severity_db)
                    session.query(Checker) \
                        .filter(Checker.id == checker_row.id) \
                        .update({Checker.severity: new_severity_db})
                    count += 1

                session.flush()

            if count:
                LOG.debug("[%s] %d checker severities upgraded.",
                          product.endpoint, count)

            session.commit()

            LOG.debug("[%s] Upgrading severity levels finished.",
                      product.endpoint)
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("[%s] Failed to upgrade severity levels: %s",
                      product.endpoint, str(ex))


def get_foreign_keys(
        session,
        table_name,
        referred_table_name,
        constraint_name=None):
    inspector = sqlalchemy.inspect(session.connection())

    foreign_keys = list(filter(
        lambda fk: fk["referred_table"] == referred_table_name
        and (fk["name"] == constraint_name if constraint_name else True),
        inspector.get_foreign_keys(table_name)
    ))
    return foreign_keys


def drop_foreign_keys(session, table_name, foreign_keys):
    for fk in foreign_keys:
        session.execute(sqlalchemy.text(
            f"ALTER TABLE {table_name} "
            f"DROP CONSTRAINT {fk['name']};"
        ))
    session.commit()


def add_foreign_keys(session, table_name, foreign_keys):
    for fk in foreign_keys:
        constraint_name = fk["name"]
        constrained_columns = ", ".join(fk["constrained_columns"])
        referred_table = fk["referred_table"]
        referred_columns = ", ".join(fk["referred_columns"])

        if get_foreign_keys(
            session,
            table_name,
            referred_table,
            constraint_name
        ):
            LOG.warning(f"Cannot add {constraint_name} constraint, "
                        "it is already exists.")
            continue

        session.execute(sqlalchemy.text(
            f"ALTER TABLE {table_name} "
            f"ADD CONSTRAINT {constraint_name} "
            f"FOREIGN KEY ({constrained_columns}) "
            f"REFERENCES {referred_table}({referred_columns});"
        ))
    session.commit()



"""
Add analyzer_name for report

Revision ID: af5d8a21c1e4
Revises:     2185167f8568
Create Date: 2020-08-14 11:26:01.806877
"""

from alembic import op
import sqlalchemy as sa

# Revision identifiers, used by Alembic.
revision = 'af5d8a21c1e4'
down_revision = '2185167f8568'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('reports', sa.Column('analyzer_name',
                                       sa.String(),
                                       server_default='unknown',
                                       nullable=False))


def downgrade():
    op.drop_column('reports', 'analyzer_name')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""

LDAP authentication module for CodeChecker.
Authenticate user based on the server_config.json LDAP part.

In the configuration `null` means it is not configured.

"method_ldap": {
  "enabled" : true,
  "authorities": [
    {
      "connection_url" : null,
      "tls_require_cert" : null,
      "username" : null,
      "password" : null,
      "referrals" : false,
      "deref" : "always",
      "accountBase" : null,
      "accountScope" : "subtree",
      "accountPattern" : null,
      "groupBase" : null,
      "groupScope" : "subtree",
      "groupPattern" : null,
      "groupNameAttr" : null
    }
  ]
},


##### Configuration options:

`connection_url`
URL of the LDAP server which will be queried for user information and group
membership.

`username`
Optional username for LDAP bind, if not set bind with the
login credentials will be attempted.

`password`
Optional password for configured username.

`referrals`
Microsoft Active Directory by returns referrals (search continuations). LDAPv3
does not specify which credentials should be used by the clients when chasing
these referrals and will be tried as an anonymous access by the libldap library
which might fail. Will be disabled by default.

`deref`
Configure how the alias dereferencing is done in libldap
(valid values: always, never).

`accountBase`
Root tree containing all the user accounts.

`accountScope`
Scope of the search performed. Accepted values are: base, one, subtree.

`accountPattern`
The special `$USN$` token in the query is replaced to the *username* at login.
Query pattern used to search for a user account. Must be a valid LDAP query
expression.
Example configuration: *(&(objectClass=person)(sAMAccountName=$USN$))*

`groupBase`
Root tree containing all the groups.

`groupPattern`

Group query pattern used LDAP query expression to find the group objects
a user is a member of. It must contain a `$USERDN$` pattern.
`$USERDN$` will be automatically replaced by the queried user account DN.

`groupNameAttr`

The attribute of the group object which contains the name of the group.

`groupScope`
Scope of the search performed. (Valid values are: base, one, subtree)


"""


from contextlib import contextmanager

import ldap
from ldap.dn import escape_dn_chars

from codechecker_common.logger import get_logger

LOG = get_logger('server')


def bytes_to_str(data):
    """ Converts the given data to text.

    The LDAP protocol states that some fields be encoded in UTF-8. Attribute
    values, on the other hand, MAY contain any type of data, including text.
    Encoding/decoding to other formats – text, images, etc. – is left to the
    caller. See:
    https://www.python-ldap.org/en/python-ldap-3.2.0/bytes_mode.html
    """
    if isinstance(data, bytes):
        try:
            return data.decode('utf-8')
        except UnicodeDecodeError as err:
            LOG.error("Failed to convert byte (%s) to text: %s", data, err)
            return ''

    return data


@contextmanager
def ldap_error_handler():
    """
    Handle LDAP errors.
    """
    try:
        yield
    except ldap.INVALID_CREDENTIALS:
        LOG.warning("Invalid credentials, please recheck "
                    "your authentication configuration.")

    except ldap.FILTER_ERROR as ex:
        LOG.error("Filter error: %s", str(ex))

    except ldap.SERVER_DOWN:
        LOG.error("Can't connect to LDAP server,"
                  " or LDAPS certificate verification failed")

    except ldap.LDAPError as err:
        LOG.error("Exception ldap.%s (%s)",
                  type(err).__name__, str(err))


def get_user_dn(con,
                account_base_dn,
                account_pattern,
                scope=ldap.SCOPE_SUBTREE,
                user_dn_postfix_preference=None):
    """
    Search for the user dn based on the account pattern.
    Return the full user dn None if search failed.

    Parameters:
      user_dn_postfix_preference: User DN postfix preference value can
      be used to select out one prefered
      user DN if multiple DN entries are found by the LDAP search.
      The configured value will be matched and the first matching will be used.
      If only one DN was found this postfix matching will not be used.
      If not set and multiple values are found the first value
      in the search result list will be used.
    """

    with ldap_error_handler():
        # Attribute values MAY contain any type of data. Before you use a
        # value, call 'bytes_to_str' helper function to convert it to text.
        user_data = con.search_s(account_base_dn, scope, account_pattern)
        user_dns = []
        if user_data:
            # User found use the user DN from the first result.
            for user_info in user_data:
                user_dns.append(bytes_to_str(user_info[0]))
            LOG.debug("Found user dns: %s", ', '.join(user_dns))

            if len(user_dns) > 1 and user_dn_postfix_preference:
                for user_dn in user_dns:
                    if user_dn.endswith(user_dn_postfix_preference):
                        LOG.debug("Selected user dn: %s", user_dn)
                        return user_dn
            elif len(user_dns) > 0:
                LOG.debug("Selected user dn: %s", user_dns[0])
                return user_dns[0]

    LOG.debug("Searching for user failed with pattern: %s", account_pattern)
    LOG.debug("Account base DN: %s", account_base_dn)
    return None


def check_group_membership(connection,
                           group_base,
                           member_query,
                           group_scope):
    """
    Check if a user is a member of an LDAP group using the memberOf
    attribute.
    """

    LOG.debug(member_query)

    with ldap_error_handler():
        # Attribute values MAY contain any type of data. Before you use a
        # value, call 'bytes_to_str' helper function to convert it to text.
        group_result = connection.search_s(group_base,
                                           group_scope,
                                           member_query,
                                           ['memberOf'])

        # There is at least one match for one of the groups.
        return len(group_result) != 0

    return False


class LDAPConnection:
    """
    A context manager class to initialize an LDAP connection
    and bind to the LDAP server.
    """

    def __init__(self, ldap_config, who=None, cred=None):
        """
        Initialize an ldap connection object and bind to the configured
        LDAP server.
        None if initialization failed.
        """

        ldap_server = ldap_config.get('connection_url')
        if ldap_server is None:
            LOG.error('Server address is missing from the configuration')
            self.connection = None
            return

        referrals = ldap_config.get('referrals', False)
        ldap.set_option(ldap.OPT_REFERRALS, 1 if referrals else 0)

        deref = ldap_config.get('deref', ldap.DEREF_ALWAYS)
        if deref == 'never':
            deref = ldap.DEREF_NEVER
        else:
            deref = ldap.DEREF_ALWAYS

        ldap.set_option(ldap.OPT_DEREF, deref)

        ldap.protocol_version = ldap.VERSION3

        # Verify certificate in LDAPS connections
        tls_require_cert = ldap_config.get('tls_require_cert', '')
        if tls_require_cert.lower() == 'never':
            LOG.debug("Insecure LDAPS connection because of "
                      "tls_require_cert=='never'")
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

        self.connection = ldap.initialize(ldap_server, bytes_mode=False)

        LOG.debug('Binding to LDAP server with user: %s', who if who else '')

        res = None
        with ldap_error_handler():
            if who is None or cred is None:
                LOG.debug("Anonymous bind with no credentials.")
                res = self.connection.simple_bind_s()
                LOG.debug(res)
            else:
                LOG.debug("Binding with credential: %s", who)
                res = self.connection.simple_bind_s(who, cred)
                whoami = self.connection.whoami_s()
                LOG.debug(res)
                LOG.debug(whoami)

                # mail.python.org/pipermail/python-ldap/2012q4/003180.html
                if whoami is None:
                    # If LDAP server allows anonymous binds, simple bind
                    # does not throw an exception when the password is
                    # empty and does the binding as anonymous.
                    # This is an expected behaviour as per LDAP RFC.

                    # However, if the bind is successful but no
                    # authentication has been done, it is still to be
                    # considered an error from the user's perspective.
                    LOG.debug("Anonymous bind succeeded but no valid "
                              "password was given.")
                    raise ldap.INVALID_CREDENTIALS()

        if not res:
            LOG.debug("Server bind failed.")
            if self.connection is not None:
                self.connection.unbind()
            self.connection = None

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, value, traceback):
        if self.connection is not None:
            self.connection.unbind()


def get_ldap_query_scope(scope_form_config):
    """
    Return an ldap scope base on the configured value.
    The defaul scope is the subtree.
    """
    # the default scope is the subtree
    if scope_form_config == 'base':
        return ldap.LDAP_SCOPE_BASE
    elif scope_form_config == 'one':
        return ldap.LDAP_SCOPE_ONELEVEL
    else:
        return ldap.SCOPE_SUBTREE


def auth_user(ldap_config, username=None, credentials=None):
    """
    Authenticate a user.
    """
    if not username or not credentials:
        LOG.warning('No username or credential is provided for'
                    ' authentication.')
        return False

    account_base = ldap_config.get('accountBase')
    if account_base is None:
        LOG.warning('Account base needs to be configured to query users')
        return False

    account_pattern = ldap_config.get('accountPattern')
    if account_pattern is None:
        LOG.warning('No account pattern is defined to search for users.')
        LOG.warning('Please configure one.')
        return False

    username = escape_dn_chars(username)
    account_pattern = account_pattern.replace('$USN$', username)

    account_scope = ldap_config.get('accountScope', '')
    account_scope = get_ldap_query_scope(account_scope)

    service_user = ldap_config.get('username')
    service_cred = ldap_config.get('password')

    # Service user is not configured try to authenticate
    # with the given username and credentials.
    if not service_user:
        service_user = username
        service_cred = credentials

    LOG.debug("Creating SERVICE connection...")

    user_dn_postfix_preference = ldap_config.get('user_dn_postfix_preference')

    with LDAPConnection(ldap_config, service_user, service_cred) as connection:
        if connection is None:
            LOG.error('Please check your LDAP server '
                      'authentication credentials.')
            LOG.error('Configured username: %s', service_user)
            return False

        user_dn = get_user_dn(connection,
                              account_base,
                              account_pattern,
                              account_scope,
                              user_dn_postfix_preference)

        if user_dn is None:
            LOG.warning("DN lookup failed for user name: '%s'!", username)
            if service_user is None:
                LOG.warning('Anonymous bind might not be enabled.')
            return False

    # Bind with the user's DN to check the password given by the user.
    # If bind is successful the user has given the right password.
    LOG.debug("Creating USER connection...")
    with LDAPConnection(ldap_config, user_dn, credentials) as connection:
        if not connection:
            LOG.info("User: %s cannot be authenticated.", username)

        return connection is not None


def get_groups(ldap_config, username, credentials):
    """
    Get the LDAP groups for a given user.
    """

    account_base = ldap_config.get('accountBase')
    if account_base is None:
        LOG.error('Account base needs to be configured to query users')
        return False

    account_pattern = ldap_config.get('accountPattern')
    if account_pattern is None:
        LOG.error('No account pattern is defined to search for users.')
        LOG.error('Please configure one.')
        return False

    account_pattern = account_pattern.replace('$USN$', username)

    account_scope = ldap_config.get('accountScope', '')
    account_scope = get_ldap_query_scope(account_scope)

    service_user = ldap_config.get('username')
    service_cred = ldap_config.get('password')
    if not service_user:
        service_user = username
        service_cred = credentials

    user_dn_postfix_preference = ldap_config.get('user_dn_postfix_preference')

    LOG.debug("creating LDAP connection. service user %s", service_user)
    with LDAPConnection(ldap_config, service_user, service_cred) as connection:
        if connection is None:
            LOG.error('Please check your LDAP server '
                      'authentication credentials.')
            return False

        user_dn = get_user_dn(connection,
                              account_base,
                              account_pattern,
                              account_scope,
                              user_dn_postfix_preference)

        group_pattern = ldap_config.get('groupPattern')
        if user_dn and not group_pattern:
            LOG.debug("User '%s' found but there is no group_pattern"
                      " to check LDAP for group membership.",
                      user_dn)
            return []
        group_pattern = group_pattern.replace('$USERDN$', user_dn)

        LOG.debug('Checking for group membership %s', user_dn)

        group_scope = ldap_config.get('groupScope', '')
        group_scope = get_ldap_query_scope(group_scope)

        group_base = ldap_config.get('groupBase')
        if group_base is None:
            LOG.error('Group base needs to be configured to'
                      'query ldap groups.')
            return []

        group_name_attr = ldap_config.get('groupNameAttr')
        if group_name_attr is None:
            LOG.error('groupNameAttr needs to be configured to'
                      'query ldap groups.'
                      'Its value must be the name'
                      'attribute of the group.')
            return []

        # Remove non ascii characters.
        group_name_attr = \
            group_name_attr.encode('ascii', 'ignore').decode('utf-8')
        attr_list = [group_name_attr]

        LOG.debug("Performing LDAP search for group: %s Group Name Attr: %s",
                  group_pattern, group_name_attr)

        groups = []
        with ldap_error_handler():
            # Attribute values MAY contain any type of data. Before you use a
            # value, call 'bytes_to_str' helper function to convert it to text.
            group_result = connection.search_s(group_base,
                                               group_scope,
                                               group_pattern,
                                               attr_list)
            if group_result:
                for g in group_result:
                    groups.append(bytes_to_str(g[1][group_name_attr][0]))

        LOG.debug("groups:")
        LOG.debug(groups)
        return groups



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Config handler for Gcc analyzer.
"""
from .. import config_handler


class GccConfigHandler(config_handler.AnalyzerConfigHandler):
    """
    Configuration handler for Gcc analyzer.
    """
    pass



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Contains functions to format and pretty-print data from two-dimensional arrays.
"""


import json

from operator import itemgetter
from typing import Iterable, List, Optional


def to_str(
    format_name: str,
    keys,
    rows,
    sort_by_column_number: Optional[int] = None,
    rev=False,
    separate_footer=False
) -> str:
    """
    Converts the given two-dimensional array (with the specified keys)
    to the given format.
    """
    if sort_by_column_number is not None:
        rows.sort(key=itemgetter(sort_by_column_number), reverse=rev)

    all_rows = rows
    if keys:
        all_rows = [keys] + list(rows)

    if format_name == 'rows':
        return to_rows(rows)
    elif format_name == 'table' or format_name == 'plaintext':
        # TODO: 'plaintext' for now to support the 'CodeChecker cmd' interface.
        return to_table(all_rows, True, separate_footer)
    elif format_name == 'csv':
        return to_csv(all_rows)
    elif format_name == 'dictlist':
        return to_dictlist(keys, rows)
    elif format_name == 'json':
        return json.dumps(to_dictlist(keys, rows))
    else:
        raise ValueError("Unsupported format")


def to_rows(lines: Iterable[str]) -> str:
    """
    Prints the given rows with minimal formatting.
    """

    str_parts = []

    lns: List[List[str]] = [
        ['' if e is None else e for e in line] for line in lines]

    # Count the column width.
    widths: List[int] = []
    for line in lns:
        for i, size in enumerate([len(str(x)) for x in line]):
            while i >= len(widths):
                widths.append(0)
            if size > widths[i]:
                widths[i] = size

    # Generate the format string to pad the columns.
    print_string = " "
    for i, width in enumerate(widths):
        if i == 0 or i == len(widths) - 1 or width == 0:
            print_string += "{" + str(i) + "} "
        else:
            print_string += "{" + str(i) + ":" + str(width) + "} "

    if not print_string:
        return ''

    print_string = print_string[:-1]

    # Print the actual data.
    for i, line in enumerate(lns):
        try:
            str_parts.append(print_string.format(*line))
        except IndexError:
            raise TypeError("One of the rows have a different number of "
                            "columns than the others")

    return '\n'.join(str_parts)


def to_table(
    lines: Iterable[str],
    separate_head=True,
    separate_footer=False
) -> str:
    """
    Pretty-prints the given two-dimensional array's lines.
    """

    str_parts = []

    # It is possible that one of the item in the line is None which will
    # raise an exception when passed to the format function below. So this is
    # the reason why we need to convert None values to valid strings here.
    lns: List[List[str]] = [
        ['' if e is None else e for e in line] for line in lines]

    # Count the column width.
    widths: List[int] = []
    for line in lns:
        for i, size in enumerate([len(str(x)) for x in line]):
            while i >= len(widths):
                widths.append(0)
            if size > widths[i]:
                widths[i] = size

    # Generate the format string to pad the columns.
    print_string = ""
    for i, width in enumerate(widths):
        print_string += "{" + str(i) + ":" + str(width) + "} | "

    if not print_string:
        return ''

    print_string = print_string[:-3]

    # Print the actual data.
    str_parts.append("-" * (sum(widths) + 3 * (len(widths) - 1)))
    for i, line in enumerate(lns):
        try:
            str_parts.append(print_string.format(*line))
        except IndexError:
            raise TypeError("One of the rows have a different number of "
                            "columns than the others")
        if i == 0 and separate_head:
            str_parts.append("-" * (sum(widths) + 3 * (len(widths) - 1)))
        if separate_footer and i == len(lns) - 2:
            str_parts.append("-" * (sum(widths) + 3 * (len(widths) - 1)))

    str_parts.append("-" * (sum(widths) + 3 * (len(widths) - 1)))

    return '\n'.join(str_parts)


def to_csv(lines: Iterable[str]) -> str:
    """
    Pretty-print the given two-dimensional array's lines in CSV format.
    """

    str_parts = []

    lns: List[List[str]] = [
        ['' if e is None else e for e in line] for line in lines]

    # Count the columns.
    columns = 0
    for line in lns:
        if len(line) > columns:
            columns = len(line)

    print_string = ""
    for i in range(columns):
        print_string += "{" + str(i) + "},"

    if not print_string:
        return ''

    print_string = print_string[:-1]

    # Print the actual data.
    for line in lns:
        try:
            str_parts.append(print_string.format(*line))
        except IndexError:
            raise TypeError("One of the rows have a different number of "
                            "columns than the others")

    return '\n'.join(str_parts)


def to_dictlist(key_list, lines):
    """
    Pretty-print the given two-dimensional array's lines into a JSON
    object list. The key_list acts as the "header" of the table, specifying the
    keys to use in the resulting object.

    This function expects values to be the same number as the length of
    key_list, and that the order of values in a line corresponds to the order
    of keys.
    """

    res = []
    for line in lines:
        res.append({key: value for (key, value) in zip(key_list, line)})

    return res



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Context to store package related information.
"""


# pylint: disable=no-name-in-module
from distutils.spawn import find_executable
from argparse import ArgumentTypeError

import os
import platform
import sys

from pathlib import Path

from codechecker_common import logger
from codechecker_analyzer.arg import analyzer_binary
from codechecker_common.checker_labels import CheckerLabels
from codechecker_common.singleton import Singleton
from codechecker_common.util import load_json

from . import env

LOG = logger.get_logger('system')


# -----------------------------------------------------------------------------
class Context(metaclass=Singleton):
    """
    Generic package specific context.

    This class is to query all information that might be important about the
    running environment around CodeChecker analysis. This is a singleton
    object, so it is cheap to construct and can be used as a read-only
    dictionary of the data on its interface.
    """

    def __init__(self):
        """ Initialize analyzer context. """
        self._bin_dir_path = os.environ.get('CC_BIN_DIR', '')
        self._lib_dir_path = os.environ.get('CC_LIB_DIR', '')
        self._data_files_dir_path = os.environ.get('CC_DATA_FILES_DIR', '')

        # Use this environment variable for testing purposes only. This
        # variable helps to configure which labels to use in this context.
        labels_dir = os.path.join(self._data_files_dir_path,
                                  'config', 'labels')
        if 'CC_TEST_LABELS_DIR' in os.environ:
            labels_dir = os.environ['CC_TEST_LABELS_DIR']

        cfg_dict = self.__get_package_config()
        self.env_vars = cfg_dict['environment_variables']

        lcfg_dict = self.__get_package_layout()
        self.pckg_layout = lcfg_dict['runtime']

        self._checker_labels = CheckerLabels(labels_dir)
        self.__package_version = None
        self.__package_build_date = None
        self.__package_git_hash = None
        self.__analyzers = {}
        self.__analyzer_env = None

        machine = platform.uname().machine

        self.logger_lib_dir_path = os.path.join(
            self._data_files_dir_path, 'ld_logger', 'lib', machine)

        if not os.path.exists(self.logger_lib_dir_path):
            self.logger_lib_dir_path = os.path.join(
                self._lib_dir_path,
                'codechecker_analyzer',
                'ld_logger',
                'lib',
                machine)

        self.logger_bin = None
        self.logger_file = None
        self.logger_compilers = None

        # Init package specific environment variables.
        self.__init_env()

        self.__set_version()
        self.__populate_analyzers()
        self.__populate_replacer()

    def __parse_CC_ANALYZER_BIN(self):
        env_var_bins = {}
        if 'CC_ANALYZER_BIN' in self.analyzer_env:
            had_error = False
            for value in self.__analyzer_env['CC_ANALYZER_BIN'].split(';'):
                try:
                    analyzer_name, path = analyzer_binary(value)
                except ArgumentTypeError as e:
                    LOG.error(e)
                    had_error = True
                    continue

                resolved_path = find_executable(path)

                if not os.path.isfile(resolved_path):
                    LOG.error(f"'{path}' is not a path to an analyzer binary "
                              "given to CC_ANALYZER_BIN!")
                    had_error = True

                if had_error:
                    continue

                LOG.info(f"Using '{resolved_path}' for analyzer "
                         f"'{analyzer_name}'")
                env_var_bins[analyzer_name] = resolved_path

            if had_error:
                LOG.info("The value of CC_ANALYZER_BIN should be in the"
                         "format of "
                         "CC_ANALYZER_BIN='<analyzer1>:/path/to/bin1;"
                         "<analyzer2>:/path/to/bin2'")
                sys.exit(1)
        return env_var_bins

    def __get_package_config(self):
        """ Get package configuration. """
        pckg_config_file = os.path.join(
            self._data_files_dir_path, "config", "config.json")

        LOG.debug('Reading config: %s', pckg_config_file)
        cfg_dict = load_json(pckg_config_file)

        if not cfg_dict:
            raise ValueError(f"No configuration file '{pckg_config_file}' can "
                             f"be found or it is empty!")

        LOG.debug(cfg_dict)
        return cfg_dict

    def __get_package_layout(self):
        """ Get package layout configuration. """
        layout_cfg_file = os.path.join(
            self._data_files_dir_path, "config", "package_layout.json")

        LOG.debug('Reading config: %s', layout_cfg_file)
        lcfg_dict = load_json(layout_cfg_file)

        if not lcfg_dict:
            raise ValueError(f"No configuration file '{layout_cfg_file}' can "
                             f"be found or it is empty!")

        return lcfg_dict

    def __init_env(self):
        """ Set environment variables. """
        # Get generic package specific environment variables.
        self.logger_bin = os.environ.get(self.env_vars['cc_logger_bin'])
        self.logger_file = os.environ.get(self.env_vars['cc_logger_file'])
        self.logger_compilers = os.environ.get(
            self.env_vars['cc_logger_compiles'])
        self.ld_preload = os.environ.get(self.env_vars['ld_preload'])
        self.ld_lib_path = self.env_vars['env_ld_lib_path']

    def __set_version(self):
        """
        Get the package version from the version config file.
        """
        vfile_data = load_json(self.version_file)

        if not vfile_data:
            sys.exit(1)

        package_version = vfile_data['version']
        package_build_date = vfile_data['package_build_date']
        package_git_hash = vfile_data.get('git_hash')
        package_git_tag = vfile_data.get('git_describe', {}).get('tag')
        package_git_dirtytag = vfile_data.get('git_describe', {}).get('dirty')

        self.__package_version = package_version['major'] + '.' + \
            package_version['minor'] + '.' + \
            package_version['revision'] + \
            ('-rc' + package_version['rc'] if 'rc' in package_version and
             package_version['rc'] else '')

        self.__package_build_date = package_build_date
        self.__package_git_hash = package_git_hash

        self.__package_git_tag = package_git_tag
        if (LOG.getEffectiveLevel() == logger.DEBUG or
                LOG.getEffectiveLevel() ==
                logger.DEBUG_ANALYZER):
            self.__package_git_tag = package_git_dirtytag

    def __populate_analyzers(self):
        """ Set analyzer binaries for each registered analyzers. """
        analyzer_env = None
        analyzer_from_path = env.is_analyzer_from_path()
        if not analyzer_from_path:
            analyzer_env = self.analyzer_env

        env_var_bin = self.__parse_CC_ANALYZER_BIN()

        compiler_binaries = self.pckg_layout.get('analyzers')
        for name, value in compiler_binaries.items():
            if name in env_var_bin:
                self.__analyzers[name] = env_var_bin[name]
                continue

            if analyzer_from_path:
                value = os.path.basename(value)

            if os.path.dirname(value):
                # Check if it is a package relative path.
                self.__analyzers[name] = os.path.join(
                    self._data_files_dir_path, value)
            else:
                env_path = analyzer_env['PATH'] if analyzer_env else None
                compiler_binary = find_executable(value, env_path)
                if not compiler_binary:
                    LOG.debug("'%s' binary can not be found in your PATH!",
                              value)
                    self.__analyzers[name] = None
                    continue

                self.__analyzers[name] = os.path.realpath(compiler_binary)

                # If the compiler binary is a simlink to ccache, use the
                # original compiler binary.
                if self.__analyzers[name].endswith("/ccache"):
                    self.__analyzers[name] = compiler_binary

    def __populate_replacer(self):
        """ Set clang-apply-replacements tool. """
        replacer_binary = self.pckg_layout.get('clang-apply-replacements')

        if os.path.dirname(replacer_binary):
            # Check if it is a package relative path.
            self.__replacer = os.path.join(self._data_files_dir_path,
                                           replacer_binary)
        else:
            self.__replacer = find_executable(replacer_binary)

    @property
    def version(self):
        return self.__package_version

    @property
    def package_build_date(self):
        return self.__package_build_date

    @property
    def package_git_hash(self):
        return self.__package_git_hash

    @property
    def package_git_tag(self):
        return self.__package_git_tag

    @property
    def version_file(self):
        return os.path.join(self._data_files_dir_path, 'config',
                            'analyzer_version.json')

    @property
    def env_var_cc_logger_bin(self):
        return self.env_vars['cc_logger_bin']

    @property
    def env_var_ld_preload(self):
        return self.env_vars['ld_preload']

    @property
    def env_var_cc_logger_file(self):
        return self.env_vars['cc_logger_file']

    @property
    def path_logger_bin(self):
        return os.path.join(self._bin_dir_path, 'ld_logger')

    @property
    def logger_lib_path(self):
        """
        Returns the absolute path to the logger library.
        """
        return str(Path(self.logger_lib_dir_path,
                        self.logger_lib_name).absolute())

    @property
    def logger_lib_name(self):
        return 'ldlogger.so'

    @property
    def path_plist_to_html_dist(self):
        return os.path.join(self._lib_dir_path, 'codechecker_report_converter',
                            'report', 'output', 'html', 'static')

    @property
    def path_env_extra(self):
        if env.is_analyzer_from_path():
            return []

        extra_paths = self.pckg_layout.get('path_env_extra', [])
        paths = []
        for path in extra_paths:
            paths.append(os.path.join(self._data_files_dir_path, path))
        return paths

    @property
    def ld_lib_path_extra(self):
        if env.is_analyzer_from_path():
            return []

        extra_lib = self.pckg_layout.get('ld_lib_path_extra', [])
        ld_paths = []
        for path in extra_lib:
            ld_paths.append(os.path.join(self._data_files_dir_path, path))
        return ld_paths

    @property
    def analyzer_env(self):
        if not self.__analyzer_env:
            self.__analyzer_env = \
                env.extend(self.path_env_extra, self.ld_lib_path_extra)
        return self.__analyzer_env

    @property
    def analyzer_binaries(self):
        return self.__analyzers

    @property
    def replacer_binary(self):
        return self.__replacer

    @property
    def data_files_dir_path(self):
        return self._data_files_dir_path

    @property
    def checker_plugin(self):
        # Do not load the plugin because it might be incompatible.
        if env.is_analyzer_from_path():
            return None

        return os.path.join(self._data_files_dir_path, 'plugin')

    @property
    def checker_labels(self):
        return self._checker_labels


def get_context():
    try:
        return Context()
    except KeyError:
        import traceback
        traceback.print_exc()
        sys.exit(1)
    except ValueError as err:
        LOG.error(err)
        sys.exit(1)



"""
'off' and 'unavailable' detection statuses

Revision ID: 3e91d0612422
Revises:     40112fd406e3
Create Date: 2018-12-19 11:16:57.107510
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '3e91d0612422'
down_revision = '40112fd406e3'
branch_labels = None
depends_on = None


table_name = 'reports'
name = 'detection_status'
tmp_name = 'tmp_' + name

old_options = ('new', 'unresolved', 'resolved', 'reopened')
new_options = old_options + ('off', 'unavailable', )

new_type = sa.Enum(*new_options, name=name)
old_type = sa.Enum(*old_options, name=name)


def upgrade():
    ctx = op.get_context()
    dialect = ctx.dialect.name

    if dialect == 'postgresql':
        # Rename the enum type what we want to change.
        op.execute('ALTER TYPE ' + name + ' RENAME TO ' + tmp_name)

        # Create the new enum.
        new_type.create(op.get_bind())

        # Alter detection status column.
        op.execute('ALTER TABLE ' + table_name + ' ALTER COLUMN ' + name +
                   ' TYPE ' + name + ' USING ' + name + '::text::' + name)

        # Drop the old enum.
        op.execute('DROP TYPE ' + tmp_name)
    elif dialect == 'sqlite':
        op.execute('PRAGMA foreign_keys=off')
        with op.batch_alter_table('reports') as batch_op:
            batch_op.alter_column(name,
                                  existing_type=old_type,
                                  type_=new_type)
        op.execute('PRAGMA foreign_keys=on')


def downgrade():
    ctx = op.get_context()
    dialect = ctx.dialect.name

    if dialect == 'postgresql':
        # Rename the enum type what we want to change.
        op.execute('ALTER TYPE ' + name + ' RENAME TO ' + tmp_name)

        # Create the new enum.
        old_type.create(op.get_bind())

        # Alter detection status column.
        op.execute('ALTER TABLE ' + table_name + ' ALTER COLUMN ' + name +
                   ' TYPE ' + name + ' USING ' + name + '::text::' + name)

        # Drop the old enum.
        op.execute('DROP TYPE ' + tmp_name)
    elif dialect == 'sqlite':
        op.execute('PRAGMA foreign_keys=off')
        with op.batch_alter_table('reports') as batch_op:
            batch_op.alter_column(name,
                                  existing_type=new_type,
                                  type_=old_type)
        op.execute('PRAGMA foreign_keys=on')



"""
Initial schema

Revision ID: 8268fc7ca7f4
Revises:     <None>
Create Date: 2017-09-18 20:57:11.098460
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '8268fc7ca7f4'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'db_version',
        sa.Column('major', sa.Integer(), nullable=False),
        sa.Column('minor', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('major', 'minor', name=op.f('pk_db_version'))
    )

    op.create_table(
        'permissions_system',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('permission', sa.Enum('SUPERUSER', name='sys_perms'),
                  nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('is_group', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_permissions_system'))
    )

    op.create_table(
        'products',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('endpoint', sa.String(), nullable=False),
        sa.Column('connection', sa.String(), nullable=False),
        sa.Column('display_name', sa.String(), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_products')),
        sa.UniqueConstraint('endpoint', name=op.f('uq_products_endpoint'))
    )

    op.create_table(
        'permissions_product',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('permission', sa.Enum('PRODUCT_ADMIN',
                                        'PRODUCT_ACCESS',
                                        'PRODUCT_STORE',
                                        name='product_perms'),
                  nullable=True),
        sa.Column('product_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('is_group', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ['product_id'], ['products.id'],
            name=op.f('fk_permissions_product_product_id_products'),
            ondelete='CASCADE', initially="IMMEDIATE", deferrable=False),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_permissions_product'))
    )


def downgrade():
    op.drop_table('permissions_product')
    op.drop_table('products')
    op.drop_table('permissions_system')
    op.drop_table('db_version')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import glob
import os


def _generate_func_map_lines(func_map_dir):
    """ Iterate over all lines of input files in random order. """
    files = glob.glob(os.path.join(func_map_dir, '*'))
    for func_map_file in files:
        with open(func_map_file, 'r',
                  encoding='utf-8', errors="ignore") as func_map:
            for line in func_map:
                yield line


def _create_global_ctu_function_map(func_map_lines):
    """ Takes iterator of individual function maps and creates a global map.

    It will keeping only unique names. We leave conflicting names out of CTU.
    A function map contains the id of a function (mangled name) and the
    originating source (the corresponding AST file) name.
    """
    mangled_to_asts = {}

    # We collect all occurences of a function name into a set.
    for line in func_map_lines:
        line = line.strip()
        # FIXME: Detect and report invalid input.
        # The format of the external function map file changed in between
        # clang-15 and clang-16, check whether this is the updated format.
        # The new file format is <Length>:<USR> <File-Path>
        if line[0].isdigit():
            length_str, _ = line.split(':', 1)
            length = int(length_str)
            sep_pos = len(length_str) + 1 + length
            mangled_name = line[0: sep_pos]
            ast_file = line[sep_pos + 1:]  # Skipping the ' ' separator
        else:  # The old file format
            mangled_name, ast_file = line.split(' ', 1)
        if mangled_name not in mangled_to_asts:
            mangled_to_asts[mangled_name] = {ast_file}
        else:
            mangled_to_asts[mangled_name].add(ast_file)

    mangled_ast_pairs = []

    for mangled_name, ast_files in mangled_to_asts.items():
        if len(ast_files) == 1:
            mangled_ast_pairs.append((mangled_name, ast_files.pop()))

    return mangled_ast_pairs


def merge(func_map_dir, output_file):
    """ Merge individual function maps into a global one.

    As the collect phase runs parallel on multiple threads, all compilation
    units are separately mapped into a temporary file in ctu_temp_fnmap_folder.
    These function maps contain the mangled names of functions and the source
    (AST generated from the source) which had them.
    These files should be merged at the end into a global map file:
    ctu_func_map_file.
    """
    func_map_lines = _generate_func_map_lines(func_map_dir)
    mangled_ast_pairs = _create_global_ctu_function_map(func_map_lines)

    # Write (mangled function name, ast file) pairs into final file.
    with open(output_file, 'w',
              encoding='utf-8', errors='ignore') as out_file:
        for mangled_name, ast_file in mangled_ast_pairs:
            out_file.write('%s %s\n' % (mangled_name, ast_file))



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Utilities for compilation database handling.
"""


import os
import shlex
from typing import Callable, Dict, List, Optional

from codechecker_common.util import load_json


# For details see
# https://gcc.gnu.org/onlinedocs/gcc-12.2.0/gcc/Overall-Options.html#Overall-Options
C_CPP_OBJC_OBJCPP_EXTS = [
    '.c', '.i', '.ii', '.m', '.mi', '.mm', '.M', '.mii',
    '.cc', '.cp', '.cxx', '.cpp', '.CPP', 'c++', '.C',
    '.hh', '.H', '.hp', '.hxx', '.hpp', '.HPP', '.h++', '.tcc'
]


# Compilation database is conventionally stored in this file. Many Clang-based
# tools rely on this file name, and CMake exports this too.
COMPILATION_DATABASE = "compile_commands.json"


def find_closest_compilation_database(path: str) -> Optional[str]:
    """
    Traverse the parent directories of the given path and find the closest
    compile_commands.json. If no JSON file exists with this name up to the root
    then None returns.
    The path of the first compilation database is returned even if it doesn't
    contain a corresponding entry for the given source file in "path".
    """
    path = os.path.abspath(path)
    root = os.path.abspath(os.sep)

    while True:
        path = os.path.dirname(path)
        compile_commands_json = os.path.join(path, COMPILATION_DATABASE)

        if os.path.isfile(compile_commands_json):
            return compile_commands_json

        if path == root:
            break


def change_args_to_command_in_comp_db(compile_commands: List[Dict]):
    """
    In CodeChecker we support compilation databases where the JSON object of a
    build action contains "file", "directory" and "command" fields. However,
    compilation databases from intercept build contain "arguments" instead of
    "command" which is a list of command-line arguments instead of the same
    command as a single string. This function make this appropriate conversion.
    """
    for cc in compile_commands:
        if 'command' not in cc:
            # TODO: shlex.join(cmd) would be more elegant after upgrading to
            # Python 3.8.
            cc['command'] = ' '.join(map(shlex.quote, cc['arguments']))
            del cc['arguments']


def find_all_compilation_databases(path: str) -> List[str]:
    """
    Collect all compilation database paths that may be relevant for the source
    files at the given path. This means compilation databases anywhere under
    this path and in the parent directories up to the root.
    """
    dirs = []

    for root, _, files in os.walk(path):
        if COMPILATION_DATABASE in files:
            dirs.append(os.path.join(root, COMPILATION_DATABASE))

    path = os.path.abspath(path)
    root = os.path.abspath(os.sep)

    while True:
        path = os.path.dirname(path)
        compile_commands_json = os.path.join(path, COMPILATION_DATABASE)

        if os.path.isfile(compile_commands_json):
            dirs.append(compile_commands_json)

        if path == root:
            break

    return dirs


def build_action_describes_file(file_path: str) -> Callable[[Dict], bool]:
    """
    Returns a function which checks whether a build action belongs to the
    given file_path. This returned function can be used for filtering
    build actions in a compilation database to find build actions belonging
    to the given source file.
    """
    def describes_file(build_action) -> bool:
        return os.path.abspath(file_path) == os.path.abspath(
            os.path.join(build_action['directory'], build_action['file']))
    return describes_file


def is_c_lang_source_file(source_file_path: str) -> bool:
    """
    A file is candidate if a build action may belong to it in some
    compilation database, i.e. it is a C/C++/Obj-C source file.
    """
    return os.path.isfile(source_file_path) and \
        os.path.splitext(source_file_path)[1] in C_CPP_OBJC_OBJCPP_EXTS


def find_build_actions_for_file(file_path: str) -> List[Dict]:
    """
    Find the corresponding compilation database belonging to the given
    source file and return a list of build actions that describe its
    compilation.
    """
    comp_db = find_closest_compilation_database(file_path)

    if comp_db is None:
        return []

    return list(filter(
        build_action_describes_file(file_path),
        load_json(comp_db)))


def gather_compilation_database(analysis_input: str) -> Optional[List[Dict]]:
    """
    Return a compilation database that describes the build of the given
    analysis_input:

    - If analysis_input is a compilation database JSON file then its entries
      return.
    - If analysis_input is a C/C++/Obj-C source file then the corresponding
      build command is found from the compilation database. Only the innermost
      compilation database is checked (see find_closest_compilation_database()
      for details).
    - If analysis_input is a directory then a compilation database with the
      build commands of all C/C++/Obj-C files in it return.

    If none of these apply (e.g. analysis_input is a Python source file which
    doesn't have a compilation database) then None returns.
    """
    def __select_compilation_database(
        comp_db_paths: List[str],
        source_file: str
    ) -> Optional[str]:
        """
        Helper function for selecting the corresponding compilation database
        for the given source file, i.e. the compilation database in the closest
        parent directory, even if it doesn't contain a build action belonging
        to this file.
        """
        comp_db_paths = list(map(os.path.dirname, comp_db_paths))
        longest = os.path.commonpath(comp_db_paths + [source_file])
        if longest in comp_db_paths:
            return os.path.join(longest, COMPILATION_DATABASE)

    # Case 1: analysis_input is a compilation database JSON file.

    build_actions = load_json(analysis_input, display_warning=False)

    if build_actions is not None:
        pass

    # Case 2: analysis_input is a C/C++/Obj-C source file.

    elif is_c_lang_source_file(analysis_input):
        build_actions = find_build_actions_for_file(analysis_input)

    # Case 3: analysis_input is a directory.

    elif os.path.isdir(analysis_input):
        compilation_database_files = \
            find_all_compilation_databases(analysis_input)

        build_actions = []

        for comp_db_file in compilation_database_files:
            comp_db = load_json(comp_db_file)

            if not comp_db:
                continue

            for ba in comp_db:
                file_path = os.path.join(ba["directory"], ba["file"])

                if os.path.commonpath([file_path, analysis_input]) \
                        == analysis_input and __select_compilation_database(
                        compilation_database_files,
                        file_path) == comp_db_file:
                    build_actions.append(ba)

    # Compilation database transformation.
    if build_actions is not None:
        change_args_to_command_in_comp_db(build_actions)

    return build_actions



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
""" CodeChecker baseline output helpers. """

import logging
from typing import Iterable, List, Set, TextIO

from codechecker_report_converter.report import Report


LOG = logging.getLogger('report-converter')


def __get_report_hashes(f: TextIO) -> List[str]:
    """ Get report hashes from the given file. """
    return [h.strip() for h in f.readlines() if h]


def check(file_path: str) -> bool:
    """ True if the given file path is a baseline file. """
    return file_path.endswith('.baseline')


def get_report_hashes(
    baseline_file_paths: Iterable[str]
) -> Set[str]:
    """ Get uniqued hashes from baseline files. """
    report_hashes = set()
    for file_path in baseline_file_paths:
        with open(file_path, mode='r', encoding='utf-8', errors="ignore") as f:
            report_hashes.update(__get_report_hashes(f))

    return report_hashes


def convert(reports: Iterable[Report]) -> List[str]:
    """ Convert the given reports to CodeChecker baseline format.

    Returns a list of sorted unique report hashes.
    """
    return sorted(set(
        r.report_hash for r in reports if r.report_hash is not None))


def write(file_path: str, report_hashes: Iterable[str]):
    """ Create a new baseline file or extend an existing one with the given
    report hashes in the given output directory. It will remove the duplicates
    and also sort the report hashes before writing it to a file.
    """
    with open(file_path, mode='a+', encoding='utf-8', errors="ignore") as f:
        f.seek(0)
        old_report_hashes = __get_report_hashes(f)
        new_report_hashes = set(report_hashes) - set(old_report_hashes)

        if not new_report_hashes:
            LOG.info("Baseline file (%s) is up-to-date.", file_path)
            return

        if old_report_hashes:
            LOG.info("Merging existing baseline file: %s", file_path)
        else:
            LOG.info("Creating new baseline file: %s", file_path)

        LOG.info("Total number of old report hashes: %d",
                 len(old_report_hashes))
        LOG.info("Total number of new report hashes: %d",
                 len(new_report_hashes))

        LOG.debug("New report hashes: %s", sorted(new_report_hashes))

        f.seek(0)
        f.truncate()
        f.write("\n".join(sorted(
            set([*old_report_hashes, *report_hashes]))))
